"""Bayesian Optimization: Select experiments from the predicted posterior and update the prior"""

from .base import Optimizer

from obsidian.parameters import ParamSpace, Target, Task
from obsidian.surrogates import SurrogateBoTorch, EnsembleModel
from obsidian.acquisition import aq_class_dict, aq_defaults, aq_hp_defaults, valid_aqs
from obsidian.surrogates import model_class_dict
from obsidian.objectives import Index_Objective, Objective_Sequence
from obsidian.constraints import Linear_Constraint, Nonlinear_Constraint, Output_Constraint
from obsidian.exceptions import IncompatibleObjectiveError, UnsupportedError, UnfitError, DataWarning, OptimizerWarning
from obsidian.config import TORCH_DTYPE

from botorch.acquisition.objective import MCAcquisitionObjective
from botorch.optim import optimize_acqf, optimize_acqf_mixed
from botorch.sampling import SobolQMCNormalSampler
from botorch.sampling.list_sampler import ListSampler
from botorch.sampling.index_sampler import IndexSampler
from botorch.models.model_list_gp_regression import ModelListGP
from botorch.models.gpytorch import GPyTorchModel
from botorch.models.model import ModelList, Model
from botorch.utils.sampling import draw_sobol_samples
from botorch.utils.multi_objective.box_decompositions.non_dominated import NondominatedPartitioning

import torch
from torch import Tensor
import pandas as pd
import numpy as np
import warnings


class BayesianOptimizer(Optimizer):
    """
    BayesianOptimizer is a class that implements a Bayesian optimization algorithm.

    This class is used to optimize a given function by iteratively selecting the next set of input parameters
    based on the results of previous evaluations. It uses a surrogate model to approximate the underlying function
    and an acquisition function to determine the next set of parameters to evaluate.

    Args:
        X_space (ParamSpace): The parameter space defining the search space for the optimization.
        surrogate (str | dict | list[str] | list[dict], optional): The surrogate model(s) to use.
            It can be a string representing a single model type, a dictionary specifying multiple model types
            with their hyperparameters, or a list of strings or dictionaries.
            
            Defaults to ``'GP'``. Options are as follows:
            
            - ``'GP'``: Gaussian Process with default settings (Matern Kernel, Gamma covariance priors)
            - ``'MixedGP'``: GP with mixed parameter types (continuous, categorical). Will be re-selected
              by default if 'GP' is selected and input space is mixed.
            - ``'DKL'``: GP with a NN feature-extractor (deep kernel learning)
            - ``'GPflat'``: GP without priors. May result in optimization instability, but removes bias
              for special situations.
            - ``'GPprior'``: GP with custom priors on the mean, likelihood, and covariance
            - ``'MTGP'``: Multi-task GP for multi-output optimization. Will be re-selected by default
              if 'GP' is selected and the input space contains Task parameters.
            - ``'DNN'``: Dropout neural network. Uses MC sampling to mask neurons during training and
              to estimate uncertainty.
            
            
        seed (int | None, optional): The random seed to use. Defaults to ``None``.
        verbose (int, optional): The verbosity level. Defaults to ``1``.

    Attributes:
        surrogate_type (list[str]): The shorthand name of each surrogate model.
        surrogate_hps (list[dict]): The hyperparameters for each surrogate model.
        is_fit (bool): Indicates whether the surrogate model has been fit to data.

    Raises:
        TypeError: If the surrogate argument is not a string, dict, or list of str/dict.
        ValueError: If the surrogate dictionary contains more than one surrogate model type.
        KeyError: If the surrogate model is not selected from the available models.
        ValueError: If the number of responses does not match the number of specified surrogate

    """
    def __init__(self,
                 X_space: ParamSpace,
                 surrogate: str | dict | list[str] | list[dict] = 'GP',
                 seed: int | None = None,
                 verbose: int = 1):
       
        super().__init__(X_space=X_space, seed=seed, verbose=verbose)

        self.surrogate_type = []  # Shorthand name as str (as provided)
        self.surrogate_hps = []  # Hyperparameters

        # Surrogate model selection
        if not isinstance(surrogate, (str, list, dict)):
            raise TypeError('Surrogate argument must be a string, list of strings \
                            dict of {surrogate: {hypers}} or list of dicts')
        
        def _load_surrogate_str(surrogate_str):
            # Just a string = use the same model type for each ycol that might appear later
            self.surrogate_type.append(surrogate_str)
            self.surrogate_hps.append(dict())

        def _load_surrogate_dict(surrogate_dict):
            if len(surrogate_dict) != 1:
                raise ValueError('Surrogate dictionary must contain only one surrogate model type')
            # Dictionary of dictionaries = hyperparameters may be included
            for surrogate_str, surrogate_hps in surrogate_dict.items():
                if not isinstance(surrogate_hps, dict):
                    raise TypeError('Surrogate dictionary must be a nested dictionary providing hyperparameters')
                self.surrogate_type.append(surrogate_str)
                self.surrogate_hps.append(surrogate_hps)

        if isinstance(surrogate, str):
            _load_surrogate_str(surrogate)
        elif isinstance(surrogate, dict):
            _load_surrogate_dict(surrogate)
        elif isinstance(surrogate, list):
            for surrogate_i in surrogate:
                if isinstance(surrogate_i, str):
                    _load_surrogate_str(surrogate_i)
                elif isinstance(surrogate_i, dict):
                    _load_surrogate_dict(surrogate_i)
                else:
                    raise ValueError('Surrogate argument must be a string, dict, or list of str/dict')
                
        for surrogate_str in self.surrogate_type:
            if surrogate_str not in model_class_dict.keys():
                raise KeyError(f'Surrogate model must be selected from one of: {model_class_dict.keys()}')

        return
    
    @property
    def is_fit(self):
        """
        Check if all surrogate mdoels in optimizer are fit

        Returns:
            bool: True if the optimizer is fit, False otherwise.
        """
        if hasattr(self, 'surrogate'):
            return all(model.is_fit for model in self.surrogate)
        else:
            return False

    def _validate_target(self,
                         target: Target | list[Target] | None = None):
        """
        Validates the target input for the optimization process.

        Args:
            target (Target | list[Target] | None, optional): The target object or a list of target objects to be validated.
                If None, the target object specified during the initialization of the optimizer will be used.
                Defaults to ``None``.

        Raises:
            TypeError: If the target is not a Target object or a list of Target objects.

        Returns:
            list[Target]: List of validated target(s)
        """
        if target is None:
            if not hasattr(self, 'target'):
                raise TypeError('Target must be a Target object or a list of Target objects')
            else:
                target = list(self.target)
        else:
            if not isinstance(target, (Target, list)):
                raise TypeError('Target must be a Target object or a list of Target objects')
            elif isinstance(target, Target):
                target = [target]
            elif isinstance(target, list):
                if not all(isinstance(t, Target) for t in target):
                    raise TypeError('Each item in target must be a Target object')
        return target

    def fit(self,
            Z: pd.DataFrame,
            target: Target | list[Target]):
        """
        Fits the BO surrogate model to data.

        Args:
            Z (pd.DataFrame): Total dataset including inputs (X) and response values (y)
            target (Target or list of Target): The responses (y) to be used for optimization,
                packed into a Target object or list thereof

        Returns:
            None. Updates the model in self.surrogate

        Raises:
            NameError: If the target is not present in the data.
            ValueError: If the number of responses does not match the number of specified surrogate models.
        """
    
        self.target = tuple(self._validate_target(target))
        self.y_names = tuple([t.name for t in self.target])
        self.n_response = len(self.target)
        
        for t in self.target:
            if t.name not in Z.columns:
                raise NameError(f"Specified target {t.name} is not present in data")

        # For multi-response, specifying one model type is OK; as this will be used for all responses
        if self.n_response > 1:
            if len(self.surrogate_type) == 1:
                self.surrogate_type *= self.n_response
                self.surrogate_hps *= self.n_response
            else:
                if self.n_response != len(self.surrogate_type):
                    raise ValueError('Number of responses does not match the number \
                                     of specified surrogate models')

        # Filter out NaN by X
        X_names = list(self.X_space.X_names)
        Z_valid = Z.copy().dropna(subset=X_names)

        # Unpack X data
        self.X_train = Z_valid[X_names]
        self.X_t_train = self.X_space.encode(self.X_train)
        # Note: Z is allowed to contain columns that are neither ycols or Xcols; these will get ignored
        # Accessing the list(tuple(names)) will enforce that the order of the columns is preserved before fitting

        # Converty y (response) data to f (target) data
        self.y_train = pd.concat([Z_valid[t.name] for t in self.target], axis=1)
        self.f_train = pd.concat([t.transform_f(Z_valid[t.name], fit=True) for t in self.target], axis=1)

        # Extract the X which achieves the best sum_f
        self.X_best_f_idx = self.f_train.sum(axis=1).idxmax()
        self.X_best_f = self.X_train.iloc[self.X_best_f_idx, :].to_frame().T

        # Instantiate and fit the model(s)
        self.surrogate = []
        for i in range(self.n_response):
            self.surrogate.append(
                SurrogateBoTorch(model_type=self.surrogate_type[i], seed=self.seed,
                                 verbose=self.verbose >= 2, hps=self.surrogate_hps[i]))
            
            # Handle response NaN values on a response-by-response basis
            f_train_i = self.f_train.iloc[:, i]
            nan_indices = np.where(f_train_i.isna().values)[0]
            X_t_train_valid = self.X_t_train.drop(nan_indices)
            f_train_i_valid = f_train_i.drop(nan_indices)
            if X_t_train_valid.shape[0] < 1:
                raise ValueError(f'No valid data points for response {self.y_names[i]}')
            if f_train_i_valid.shape[0] < 1:
                raise ValueError(f'No valid response data points for response {self.y_names[i]}')
            
            # Fit the model for each response
            self.surrogate[i].fit(X_t_train_valid, f_train_i_valid,
                                  cat_dims=self.X_space.X_t_cat_idx, task_feature=self.X_space.X_t_task_idx)
            
            if self.verbose >= 1:
                print(f'{self.surrogate_type[i]} model has been fit to data'
                      + f' with an R2-train-score of: {self.surrogate[i].r2_score:.3g}'
                      + (f' and a training-loss of: {self.surrogate[i].loss:.3g}' if self.verbose >= 2 else '')
                      + f' for response: {self.y_names[i]}')
        return
    
    def save_state(self) -> dict:
        """
        Saves the parameters of the Bayesian Optimizer so that they can be reloaded without fitting.

        Returns:
            dict: A dictionary containing the fit parameters for later loading.
        
        Raises:
            UnfitError: If the surrogate model has not been fit before saving the optimizer.
        """

        if not self.is_fit:
            raise UnfitError('Surrogate model must be fit before saving optimizer')

        # Prepare a dictionary to describe the state
        config_save = {'opt_attrs': {},
                       'X_space': self.X_space.save_state(),
                       'surrogate_spec': [{func: hps} for func, hps in zip(self.surrogate_type, self.surrogate_hps)],
                       'target': [t.save_state() for t in self.target]}

        # Select some optimizer attributes to save directly
        opt_attrs = ['X_train', 'y_train',
                     'y_names', 'n_response',
                     'seed', 'X_best_f_idx', 'X_best_f']
        
        for attr in opt_attrs:
            if isinstance(getattr(self, attr), (pd.Series, pd.DataFrame)):
                config_save['opt_attrs'][attr] = getattr(self, attr).to_dict()
            else:
                config_save['opt_attrs'][attr] = getattr(self, attr)

        # Unpack the fit parameters of each surrogate model, if present
        if self.surrogate:
            model_states = []
            # Save each surrogate model using surrogate.save() methods
            for surrogate in self.surrogate:
                model_states.append(surrogate.save_state())

            config_save['model_states'] = model_states

        return config_save
    
    def __repr__(self):
        return f'BayesianOptimizer(X_space={self.X_space}, surrogate={self.surrogate_type}, target={getattr(self, "target", None)})'

    @classmethod
    def load_state(cls,
                   config_save: dict):
        """
        Loads the parameters of the Bayesian Optimizer from a previously fit optimizer.

        Args:
            config_save (dict): A dictionary containing the fit parameters for later loading.

        Returns:
            None. Updates the parameters of the BayesianOptimizer and its surrogate model.

        Raises:
            ValueError: If the number of saved models does not match the number of named models.
        """

        new_opt = cls(X_space=ParamSpace.load_state(config_save['X_space']),
                      surrogate=config_save['surrogate_spec'])
        new_opt.target = [Target.load_state(t) for t in config_save['target']]

        # Directly unpack all of the entries in opt_attrs
        for k, v in config_save['opt_attrs'].items():
            setattr(new_opt, k, v)
  
        # Unpack and encode/transform the data objects if present
        data_objects = ['X_train', 'y_train', 'X_best_f']
        if all(hasattr(new_opt, attr) for attr in data_objects):
            new_opt.X_train = pd.DataFrame(new_opt.X_train)
            new_opt.X_t_train = new_opt.X_space.encode(new_opt.X_train)
            new_opt.y_train = pd.DataFrame(new_opt.y_train, columns=new_opt.y_names)
            new_opt.X_best_f = pd.DataFrame(new_opt.X_best_f)
            
            f_train = pd.DataFrame()
            for t, y in zip(new_opt.target, new_opt.y_train.columns):
                f = t.transform_f(new_opt.y_train[y], fit=True)
                f_train = pd.concat([f_train, f.to_frame()], axis=1)
            new_opt.f_train = f_train
            
        # Unpack the models and parameteres if present
        if 'model_states' in config_save:
            if len(new_opt.surrogate_type) != len(config_save['model_states']):
                raise ValueError('The number of saved models does not match the number of named models')
            
            # Reload each surrogate model using surrogate.load() methods
            new_opt.surrogate = []
            for obj_dict in config_save['model_states']:
                new_opt.surrogate.append(SurrogateBoTorch.load_state(obj_dict))
        
        return new_opt
    
    def predict(self,
                X: pd.DataFrame,
                return_f_inv: bool = True,
                PI_range: float = 0.7) -> pd.DataFrame:
        """
        Predicts a response over a range of experiments using the surrogate function.

        Args:
            X (pd.DataFrame): Experiments to predict over.
            return_f_inv (bool, optional): Whether or not to return the inverse-transformed objective function,
                which is the raw response (unscored). The default is ``True``. Most internal calls set to ``False`` to handle
                the transformed objective function.
            PI_range (float, optional): The nominal coverage range for the returned prediction interval

        Returns:
            pd.DataFrame: Mean prediction and prediction interval for each response

        Raises:
            TypeError: If the input is not a DataFrame.
            UnfitError: If the surrogate model has not been fit before predicting.
            ValueError: If the prediction interval range is greater than 1.
            NameError: If the input does not contain all of the required predictors from the training set.
        """

        if not isinstance(X, pd.DataFrame):
            raise TypeError('X must be pd.DataFrame')
        if not self.is_fit:
            raise UnfitError('Surrogate model must be fit before predicting')
        if PI_range >= 1:
            raise ValueError('Prediction interval range must be < 1 \
                              (100% coverage of prob. density func.)')
        if not all(col in X.columns for col in self.X_train.columns):
            raise NameError('X for prediction does not contain all of the \
                            required predictors from the training set')
        
        if self.verbose >= 3:
            print(f'Predicting {X.shape[0]} experiments [...]')
        
        X_names = list(self.X_space.X_names)
        X_pred = X[X_names].dropna(subset=X_names)  # Reinforce order and non-nan before proceeding
        nan_indices = np.where(pd.isnull(X[X_names]).any(axis=1))[0].tolist()
        if nan_indices:
            warnings.warn(f'NaN values in X_pred filtered out at indices: {nan_indices}', DataWarning)

        # Scale and encode X
        X_t = self.X_space.encode(X_pred)

        preds = pd.DataFrame()
        for i in range(self.n_response):
            mu, sd = self.surrogate[i].predict(X_t)  # Returns pd.DataFrame/Series objects
            target_i = self.target[i]
            _, lb = self.surrogate[i].predict(X_t, q=(1-PI_range)/2)
            _, ub = self.surrogate[i].predict(X_t, q=1-(1-PI_range)/2)
            name = self.y_names[i]
            if return_f_inv:
                mu = target_i.transform_f(mu, inverse=True).rename(name+' (pred)')
                lb = target_i.transform_f(lb, inverse=True).rename(name+' lb')
                ub = target_i.transform_f(ub, inverse=True).rename(name+' ub')
            else:
                mu = pd.Series(mu, name=name+'_t (pred)')
                lb = pd.Series(lb, name=name+'_t lb')
                ub = pd.Series(ub, name=name+'_t ub')
            predict_i = pd.concat([mu, lb, ub], axis=1)
            preds = pd.concat([preds, predict_i], axis=1)
            
        return preds
       
    def _validate_hypers(self,
                         o_dim: int,
                         acquisition: str | dict,) -> tuple[dict, dict]:
        """
        Validates the acquisition functions and their hyperparameters.

        Args:
            o_dim (int): The output dimensionality of the final objective function
            acquisition (str | dict): Acquisition function name (str) or dictionary
                containing the acquisition function name and its hyperparameters.

        Returns:
            tuple[dict, dict]. Validated acquisition functions and hyperparmeters

        Raises:
            ValueError: If the number of hyperparameters does not match the number of acquisition functions.
            TypeError: If the hyperparameters are not provided as a dictionary.
            ValueError: If an unknown hyperparameter is passed for an acquisition function
            UnsupportedError: If the acquisition function is not supported for the optimization type.
            ValueError: If any hyperparameters are required but not provided.

        """
        # If the item is a string, use default hypers starting with an empty dict
        aq_hps = {}
        if isinstance(acquisition, str):
            aq_str = acquisition
            hps = {}
        # If the item is a dict, the structure is {name: {hypers}}
        else:
            # Validate the hypers if provided
            if len(acquisition.items()) != 1:
                raise ValueError('One dictionary of hyperparameters \
                                    must be provided for each acquisition function')
            aq_str, hps = list(acquisition.items())[0]
            if not isinstance(hps, dict):
                raise TypeError('Hyperparameters must be provided as a dictionary')
            if not all(key in aq_hp_defaults[aq_str].keys() for key in hps.keys()):
                raise ValueError('Unknown hyperpameter amongst {hps.keys()} \
                                    selected for {aq_str}, select from one of \
                                    {aq_hp_defaults[aq_str].keys()}')
            aq_hps.update(hps)

        optim_type = 'single' if o_dim == 1 else 'multi'

        # Validate the acquisition name
        if aq_str not in valid_aqs[optim_type]:
            raise UnsupportedError(f'Each acquisition function must be \
                                    selected from: {valid_aqs[optim_type]}')

        # Fill in empty hypers with defaults, as appropriate
        for key, defaults in aq_hp_defaults[aq_str].items():
            if hps.get(key) is None:
                if not defaults['optional']:
                    raise ValueError(f'Must specify hyperpameter value {key} for {aq_str}')
                if key in ['weights', 'scalarization_weights']:
                    aq_hps[key] = [1] * o_dim
                else:
                    aq_hps[key] = defaults['val']

        return (aq_str, aq_hps)

    def _parse_aq_kwargs(self,
                         aq: str,
                         hps: dict,
                         m_batch: int,
                         target_locs: list[int],
                         model: Model,
                         X_t_pending: Tensor | None = None,
                         objective: MCAcquisitionObjective | None = None) -> dict:
        """
        Parses the acquisition function keyword arguments based on the selected acquisition
        function and hyperparameters.

        Args:
            aq (str): The name of the acquisition function.
            hps (dict): The hyperparameters for the acquisition function.
            target_locs (list[int]): Indices of trained targets to use for objective function
            X_t_pending (Tensor, optional): Suggested points yet to be run
            objective (GenericMCMultiOutputObjective or GenericMCObjective, optional):
                The objective used foroptimization after calling the target models.
                Default is ``None``.

        Returns:
            dict: The parsed acquisition function keyword arguments.
        """
        
        aq_kwargs = {}

        # Establish baseline X from training and pending
        X_train = torch.tensor(self.X_space.encode(self.X_train).values, dtype=TORCH_DTYPE)
        if X_t_pending is not None:
            X_baseline = torch.concat([X_train, X_t_pending], axis=0)
        else:
            X_baseline = X_train

        # Calculate the performance on baseline X
        f_all = []
        for i in target_locs:
            X_b = pd.DataFrame(X_baseline.numpy(),
                               columns=[col for col in self.X_t_train.columns
                                        if col not in self.X_space.X_task])
            f_i, _ = self.surrogate[i].predict(X_b)
            f_all.append(f_i)
        f_t = torch.stack(f_all, axis=1)

        # If using an objective, want to calculate EI/PI from here
        o = f_t if not objective else objective(f_t.unsqueeze(0), X_baseline).squeeze(0)
        if objective:
            aq_kwargs['objective'] = objective
        
        # Improvement aqs based on inflation or deflation of best point
        if aq in ['EI', 'PI']:
            o_max = o.max(dim=0).values * (1+hps['inflate'])
            aq_kwargs.update({'best_f': o_max})
        
        # UCB based on: mu + sqrt(beta) * sqrt(variance) = mu + sqrt(beta) * sd
        if aq == 'UCB':
            aq_kwargs['beta'] = hps['beta']

        # Noisy aqs require X_train reference
        if aq in ['NEI', 'NEHVI', 'NParEGO']:
            aq_kwargs['X_baseline'] = X_baseline
            if any(isinstance(m, EnsembleModel) for m in model.models):
                aq_kwargs['cache_root'] = False

        # Hypervolume requires reference point
        if aq in ['EHVI', 'NEHVI']:

            # The reference point must be in the objective space, by default
            # use the minimum point - 10% of the range
            ref_point = hps['ref_point']
            if ref_point is None:
                max = o.max(dim=0).values
                min = o.min(dim=0).values
                ref_point = min - 0.1 * (max - min)
            else:
                ref_point = torch.tensor(ref_point)

            aq_kwargs['ref_point'] = ref_point
            
        if aq in ['NParEGO', 'NEHVI']:
            aq_kwargs['prune_baseline'] = True
            
        if aq == 'EHVI':
            aq_kwargs['partitioning'] = NondominatedPartitioning(aq_kwargs['ref_point'], Y=o)

        if aq == 'NIPV':
            X_bounds = torch.tensor([[0.0, 1.0]]*self.X_space.n_tdim, dtype=TORCH_DTYPE).T
            qmc_samples = draw_sobol_samples(bounds=X_bounds, n=128, q=m_batch)
            aq_kwargs['mc_points'] = qmc_samples.squeeze(-2)
            aq_kwargs['sampler'] = None
            if objective:
                raise UnsupportedError('NIPV does not support objectives')

        if aq == 'NParEGO':
            w = hps['scalarization_weights']
            if isinstance(w, list):
                w = torch.tensor(w)
                w = w/torch.sum(torch.abs(w))
            aq_kwargs['scalarization_weights'] = w

        if aq == 'SF':
            aq_kwargs['X_baseline'] = X_baseline

        return aq_kwargs

    def suggest(self,
                m_batch: int = 1,
                target: Target | list[Target] | None = None,
                acquisition: list[str] | list[dict] = None,
                optim_sequential: bool = True,
                optim_samples: int = 512,
                optim_restarts: int = 10,
                objective: MCAcquisitionObjective | None = None,
                out_constraints: Output_Constraint | list[Output_Constraint] | None = None,
                eq_constraints: Linear_Constraint | list[Linear_Constraint] | None = None,
                ineq_constraints: Linear_Constraint | list[Linear_Constraint] | None = None,
                nleq_constraints: Nonlinear_Constraint | list[Nonlinear_Constraint] | None = None,
                task_index: int = 0,
                fixed_var: dict[str: float | str] | None = None,
                X_pending: pd.DataFrame | None = None,
                eval_pending: pd.DataFrame | None = None) -> tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
        """
        Suggest future experiments based on a maximization of some acquisition
        function calculated from the expectation of a surrogate model.

        Args:
            m_batch (int, optional): The number of experiments to suggest at once. The default is ``1``.
            target (Target or list of Target, optional): The response(s) to be used for optimization,
            acquisition (list of str or list of dict, optional): Indicator for the desired acquisition function(s).
                A list will propose experiments for each acquisition function based on ``optim_sequential``.
                
                The default is ``['NEI']`` for single-output and ``['NEHVI']`` for multi-output.
                Options are as follows:
                
                - ``'EI'``: Expected Improvement (relative to best of ``y_train``). Accepts hyperparameter
                  ``'inflate'``, a positive or negative float to inflate/deflate the best point for explore/exploit.
                - ``'NEI'``: Noisy Expected Improvement. More robust than ``EI`` and uses all of ``y_train``,
                  but accepts no hyperparameters.
                - ``'PI'``: Probability of Improvement (relative to best of ``y_train``). Accepts hyperparameter
                  ``'inflate'``, a positive or negative float to inflate/deflate the best point for explore/exploit.
                - ``'UCB'``: Upper Confidence Bound. Accepts hyperparameter ``'beta'``, a positive float which sets
                  the number of standard deviations above the mean.
                - ``'SR'``: Simple Regret
                - ``'RS'``: Random Sampling
                - ``'Mean'``: Mean of the posterior distribution (pure exploitation/maximization of objective)
                - ``'SF'``: Space Filling. Requests points that maximize the minimumd distance to ``X_train`` based
                  on Euclidean distance.
                - ``'NIPV'``: Negative Integrated Posterior Variance. Requests the point which most improves the prediction
                  interval for a random selection of points in the design space. Used for active learning.
                - ``'EHVI'``: Expected Hypervolume Improvement. Can accept a ``ref_point``, otherwise a point just
                  below the minimum of ``y_train``.
                - ``'NEHVI'``: Noisy Expected Hypervolume Improvement. Can accept a ``ref_point``, otherwise a point
                  just below the minimum of ``y_train``.
                - ``'NParEGO'``: Noisy Pareto Efficient Global Optimization. Can accept ``scalarization_weights``, a
                  list of weights for each objective.
                
            optim_sequential (bool, optional): Whether or not to optimize batch designs sequentially
                (by fantasy) or simultaneously. Default is ``True``.
            optim_samples (int, optional): The number of samples to use for quasi Monte Carlo sampling
                of the acquisition function. Also used for initializing the acquisition optimizer.
                The default value is ``512``.
            optim_restarts (int, optional): The number of restarts to use in the global optimization
                of the acquisition function. The default value is ``10``.
            objective (MCAcquisitionObjective, optional): The objective function to be used for optimization.
                The default is ``None``.
            out_constraints (Output_Constraint | list[Output_Constraint], optional): An output constraint, or a list
                thereof, restricting the search space by outcomes. The default is ``None``.
            eq_constraints (Linear_Constraint | list[Linear_Constraint], optional): A linear constraint, or a list
                thereof, restricting the search space by equality (=). The default is ``None``.
            ineq_constraints (Linear_Constraint | list[Linear_Constraint], optional):  A linear constraint, or a list
                thereof, restricting the search space by inequality (>=). The default is ``None``.
            nleq_constraints (Nonlinear_Constraint | list[Nonlinear_Constraint], optional):  A nonlinear constraint,
                or a list thereof, restricting the search space by nonlinear feasibility. The default is ``None``.
            task_index (int, optional): The index of the task to optimize for multi-task models. The default is ``0``.
            fixed_var (dict(str:float), optional): Name of a variable and setting, over which the
                suggestion should be fixed. Default values is ``None``
            X_pending (pd.DataFrame, optional): Experiments that are expected to be run before the next optimal set
            eval_pending (pd.DataFrame, optional): Acquisition values associated with X_pending

        Returns:
            tuple[pd.DataFrame, pd.DataFrame] = (X_suggest, eval_suggest)
                X_suggest (pd.DataFrame): Experiment matrix of real input variables,
                    selected by optimizer.
                eval_suggest (pd.DataFrame): Mean results (response, prediction interval, f(response), obj
                    function for
                    each suggested experiment.
        
        Raises:
            UnfitError: If the surrogate model has not been fit before suggesting new experiments.
            TypeError: If the target is not a Target object or a list of Target objects.
            IncorrectObjectiveError: If the objective does not successfully execute on a sample.
            TypeError: If the acquisition is not a list of strings or dictionaries.
            UnsupportedError: If the provided acquisition function does not support output constraints.

        """

        if not self.is_fit:
            raise UnfitError('Surrogate model must be fit before suggesting new experiments')
            
        if self.verbose >= 2:
            print(f'Optimizing {m_batch} experiments [...]')
        
        # Use indexing to handle if suggestions are made for a subset of fit targets/surrogates
        target = self._validate_target(target)
        target_locs = [self.y_names.index(t.name) for t in target]
        
        # Select the model(s) to use for optimization
        model_list = [one_surrogate.torch_model for i, one_surrogate in enumerate(self.surrogate) if i in target_locs]
        if all(isinstance(m, GPyTorchModel) for m in model_list):
            model = ModelListGP(*model_list)
        else:
            model = ModelList(*model_list)

        # Make sure that model/objective outputs match the input requirements for aqs
        # In order to determine the number of outputs, considering objectives,
        # just call the objective on a random sample, and check the output dims
        if objective:
            try:
                X_sample = self.X_train.iloc[0, :].to_frame().T
                eval_suggest = self.evaluate(X_sample, target=target, objective=objective)
                o_dim = len([col for col in eval_suggest.columns if 'Objective' in col])
            except Exception:
                raise IncompatibleObjectiveError('Objective(s) did not successfully execute on sample')
        else:
            o_dim = len(target_locs)

        optim_type = 'single' if o_dim == 1 else 'multi'

        # Default if no aq method is provided
        if not acquisition:
            acquisition = [aq_defaults[optim_type]]

        # Type check for acquisition
        if not isinstance(acquisition, list):
            raise TypeError('acquisition must be a list of strings or dictionaries')
        if not all(isinstance(item, (str, dict)) for item in acquisition):
            raise TypeError('Each item in acquisition list must be either a string or a dictionary')
        
        # Compute static variable inputs
        fixed_features_list = self._fixed_features(fixed_var)
        if len(fixed_features_list) > 25:
            warnings.warn(f'The combinations of discrete features is large at {len(fixed_features_list)}.'
                          + ' Optimization will proceed very slowly due to the combinatorial explosion.'
                          + ' Recommend reducing the number of discrete parameters used.', OptimizerWarning)
        
        # Set up the sampler, for MC-based optimization of acquisition functions
        if not isinstance(model, ModelListGP):
            samplers = []
            for m in model.models:
                if isinstance(m, EnsembleModel):
                    sampler_i = IndexSampler(sample_shape=torch.Size([optim_samples]), seed=self.seed)
                else:
                    sampler_i = SobolQMCNormalSampler(sample_shape=torch.Size([optim_samples]), seed=self.seed)
                samplers.append(sampler_i)
            sampler = ListSampler(*samplers)
        else:
            sampler = SobolQMCNormalSampler(sample_shape=torch.Size([optim_samples]), seed=self.seed)
            
        # Calculate search bounds for optimization
        X_bounds = torch.tensor(self.X_space.search_space.values, dtype=TORCH_DTYPE)
        
        # Set up master lists to hold the candidates from multi-acquisition results
        candidates_all = []
        eval_suggest = pd.DataFrame()

        # Incorporate previously suggested X values, if provided
        if X_pending is not None:
            m_pending = X_pending.shape[0]
            candidates_pending = torch.tensor(self.X_space.encode(X_pending).values)
            candidates_all.append(candidates_pending)
            X_t_pending = torch.concat(candidates_all)
            if eval_pending is None:
                eval_suggest['aq Method'] = ['User Provided']*m_pending
            else:
                eval_suggest = eval_pending
        else:
            X_t_pending = None

        #  Select the task to optimize for a multi-task model
        if self.X_space.X_task:
            if objective is not None:
                objective = Objective_Sequence([Index_Objective(task_index), objective])
            task_param = next(x for x in self.X_space if isinstance(x, Task))
            task_name = task_param.name
            task_value = task_param.encode([task_param.categories[task_index]])

        # Proceed with the optimization over each set of acquisition/hypers
        for aq_i in acquisition:
            # Extract acq function names and custom hyperparameters from the 'acquisition' list in config
            aq_str, aq_hps = self._validate_hypers(o_dim, aq_i)

            # Use aq_kwargs so that extra unnecessary ones in hps get removed for certain aq funcs
            aq_kwargs = {'model': model, 'sampler': sampler, 'X_pending': X_t_pending}
            
            aq_kwargs.update(self._parse_aq_kwargs(aq_str, aq_hps, m_batch,
                                                   target_locs, model,
                                                   X_t_pending, objective))

            # Raise errors related to certain constraints
            if aq_str in ['UCB', 'Mean', 'TS', 'SF', 'SR', 'NIPV']:
                if out_constraints is not None:
                    raise UnsupportedError('Provided aquisition function does not support output constraints')
            else:
                if out_constraints and not isinstance(out_constraints, list):
                    out_constraints = [out_constraints]
                aq_kwargs['constraints'] = [c.forward(scale=objective is None)
                                            for c in out_constraints] if out_constraints else None

            # If NoneType, coerce to list
            if not eq_constraints:
                eq_constraints = []
            if not ineq_constraints:
                ineq_constraints = []
            if not nleq_constraints:
                nleq_constraints = []

            # Coerce input constraints to lists
            if not isinstance(eq_constraints, list):
                eq_constraints = [eq_constraints]
            if not isinstance(ineq_constraints, list):
                ineq_constraints = [ineq_constraints]
            if not isinstance(nleq_constraints, list):
                nleq_constraints = [nleq_constraints]

            # Append X_space constraints
            if getattr(self.X_space, 'linear_constraints', []):
                for c in self.X_space.linear_constraints:
                    if c.equality:
                        eq_constraints.append(c)
                    else:
                        ineq_constraints.append(c)
            if getattr(self.X_space, 'nonlinear_constraints', []):
                nleq_constraints += self.X_space.nonlinear_constraints

            # Input constraints are used by optim_acqf and friends
            optim_kwargs = {'equality_constraints': [c() for c in eq_constraints] if eq_constraints else None,
                            'inequality_constraints': [c() for c in ineq_constraints] if ineq_constraints else None,
                            'nonlinear_inequality_constraints': [c() for c in nleq_constraints] if nleq_constraints else None}
            
            optim_options = {}  # Can optionally specify batch_limit or max_iter
            
            # If nonlinear constraints are used, BoTorch doesn't provide an ic_generator
            # Must provide manual samples = just use random initialization
            if nleq_constraints:
                X_ic = torch.ones((optim_samples, 1 if fixed_features_list else m_batch, self.X_space.n_tdim))*torch.rand(1)
                optim_kwargs['batch_initial_conditions'] = X_ic
                if fixed_features_list:
                    raise UnsupportedError('Nonlinear constraints are not supported with discrete features.')
            
            # Hypervolume aqs fail with X_t_pending when optim_sequential=True
            if aq_str in ['NEHVI', 'EHVI']:
                if optim_sequential and X_t_pending is not None:
                    warnings.warn('Hypervolume aqs with X_pending require joint optimization. \
                                   Setting optim_sequential to False', UserWarning)
                    optim_sequential = False

            # If it's random search, no need to do optimization; Otherwise, initialize the aq function and optimize
            if aq_str == 'RS':
                candidates = torch.rand((m_batch, self.X_space.n_tdim), dtype=TORCH_DTYPE)
            else:
                aq_func = aq_class_dict[aq_str](**aq_kwargs).to(TORCH_DTYPE)
                
                # If there are any discrete values, we must used the mixed integer optimization
                if fixed_features_list:
                    candidates, _ = optimize_acqf_mixed(acq_function=aq_func, bounds=X_bounds,
                                                        fixed_features_list=fixed_features_list,
                                                        q=m_batch,  # Always sequential
                                                        num_restarts=optim_restarts, raw_samples=optim_samples,
                                                        options=optim_options,
                                                        **optim_kwargs)
                else:
                    candidates, _ = optimize_acqf(acq_function=aq_func, bounds=X_bounds,
                                                  q=m_batch,
                                                  sequential=optim_sequential,
                                                  num_restarts=optim_restarts, raw_samples=optim_samples,
                                                  options=optim_options,
                                                  **optim_kwargs)
            
            if self.verbose >= 2:
                print(f'Optimized {aq_str} acquisition function successfully')
            
            candidates_i = self.X_space.decode(
                pd.DataFrame(candidates.detach().cpu().numpy(),
                             columns=[col for col in self.X_t_train.columns if col not in self.X_space.X_task]))
            
            if X_t_pending is not None:
                X_t_pending_i = X_t_pending
            else:
                X_t_pending_i = None
            
            eval_i = self.evaluate(candidates_i, X_t_pending_i,
                                   target=target, acquisition=aq_i, objective=objective, eval_aq=True)
            eval_suggest = pd.concat([eval_suggest, eval_i], axis=0).reset_index(drop=True)

            candidates_all.append(candidates)
            X_t_pending = torch.concat(candidates_all)
                       
        candidates_all = pd.DataFrame(X_t_pending.detach().cpu().numpy(),
                                      columns=[col for col in self.X_t_train.columns
                                               if col not in self.X_space.X_task])
    
        if self.X_space.X_task:
            candidates_all[task_name] = task_value
        
        # Arrange suggested experiments into a single dataframe, and compute predictions
        X_suggest = self.X_space.decode(candidates_all)
        
        return X_suggest, eval_suggest
    
    def evaluate(self,
                 X_suggest: pd.DataFrame,
                 X_t_pending: Tensor | None = None,
                 target: Target | list[Target] | None = None,
                 acquisition: str | dict = None,
                 objective: MCAcquisitionObjective | None = None,
                 eval_aq: bool = False) -> pd.DataFrame:
        """
        Args:
            X_suggest (pd.DataFrame): Experiment matrix of real input variables, selected by optimizer.
            X_t_pending (Tensor): Suggested experiments yet to be run
            target (Target or list of Target, optional): The response(s) to be used for optimization,
            acquisition (str | dict, optional): Acquisition function name (str) or dictionary
                containing the acquisition function name and its hyperparameters.
            objective (MCAcquisitionObjective, optional): The objective function to be used for optimization.
                The default is ``None``.
            eval_aq (bool, optional): Whether or not to also evaluate the aq function. The default is ``False``.
        
        Returns:
            pd.DataFrame: Response prediction, pred interval, transformed mean, aq value,
                and objective function evaluation(s)

        """
        
        if not self.is_fit:
            raise UnfitError('Surrogate model must be fit before evaluating new experiments')
                
        # Use indexing to handle if suggestions are made for a subset of fit targets/surrogates
        target = self._validate_target(target)
        target_locs = [self.y_names.index(t.name) for t in target]

        # Begin evaluation with y_predict with pred interval
        eval_suggest = self.predict(X_suggest)
        X_t = torch.tensor(self.X_space.encode(X_suggest).values, dtype=TORCH_DTYPE)
        X_t_train = torch.tensor(self.X_space.encode(self.X_train).values, dtype=TORCH_DTYPE)

        # Evaluate f_predict on new and pending points
        f_all = []
        f_train_all = []
        f_pending_all = []
        for loc in target_locs:
            # Predict: new suggestions
            t_model = self.surrogate[loc]
            mu_i, _ = t_model.predict(self.X_space.encode(X_suggest))
            f_all.append(mu_i.unsqueeze(1))
            eval_suggest[f'f({self.target[loc].name})'] = mu_i

            # Training data on select targets
            mu_i_train, _ = t_model.predict(self.X_space.encode(self.X_train))
            f_train_all.append(mu_i_train.unsqueeze(1))

            # Pending points if provided
            if X_t_pending is not None:
                mu_i_pending, _ = t_model.predict(pd.DataFrame(X_t_pending,
                                                               columns=[col for col in self.X_t_train.columns
                                                                        if col not in self.X_space.X_task]))
                f_pending_all.append(mu_i_pending.unsqueeze(1))

        f_t = torch.concat(f_all, dim=1)
        f_t_train = torch.concat(f_train_all, dim=1)
        if X_t_pending is not None:
            f_t_pending = torch.concat(f_pending_all, dim=1)

        if objective:
            # Convert f_predict to tensor and evaluate objective
            # Must add sample dimension to f_t
            o = objective(f_t.unsqueeze(0), X_t).squeeze(0)
            if o.ndim < 2:
                o = o.unsqueeze(1)  # Rearrange into m x o

            # Store multiple objectives if applicable
            for o_i in range(o.shape[-1]):
                eval_suggest[f'Objective {o_i+1}'] = o[:, o_i].detach().cpu().numpy()

            # Also calculate for training and pending data (for total hv and pareto calcs)
            o_train = objective(f_t_train.unsqueeze(0), X_t_train).squeeze(0)
            if o_train.ndim < 2:
                o_train = o_train.unsqueeze(1)
            if X_t_pending is not None:
                o_pending = objective(f_t_pending.unsqueeze(0), X_t_pending).squeeze(0)
                if o_pending.ndim < 2:
                    o_pending = o_pending.unsqueeze(1)

            # Calculate output dimensionality and evaluate acquisition
            o_dim = o.shape[-1]
            
        else:
            o_dim = len(target_locs)

        optim_type = 'single' if o_dim == 1 else 'multi'
        
        if eval_aq:
            # Default if no aq method is provided
            if not acquisition:
                acquisition = [aq_defaults[optim_type]]

            if not isinstance(acquisition, (str, dict)):
                raise TypeError('Acquisition must be either a string or a dictionary')
            
            model_list = [one_surrogate.torch_model for i, one_surrogate in enumerate(self.surrogate) if i in target_locs]
            if all(isinstance(m, GPyTorchModel) for m in model_list):
                model = ModelListGP(*model_list)
            else:
                model = ModelList(*model_list)
            
            # Extract acq function names and custom hyperparameters from the 'acquisition' list in config
            aq_str, aq_hps = self._validate_hypers(o_dim, acquisition)

            # Use aq_kwargs so that extra unnecessary ones in hps get removed for certain aq funcs
            aq_kwargs = {'model': model, 'sampler': None, 'X_pending': X_t_pending}
                       
            aq_kwargs.update(self._parse_aq_kwargs(aq_str, aq_hps, X_suggest.shape[0],
                                                   target_locs, model,
                                                   X_t_pending, objective))
                
            # If it's random search, no need to evaluate aq
            if aq_str == 'RS':
                a_joint = torch.tensor([float('nan')]).repeat(X_t.shape[0]).unsqueeze(1)
            else:
                aq_func = aq_class_dict[aq_str](**aq_kwargs)

                # Evaluate acquisition on individual samples, then jointly
                a = []
                for x_i in X_t:
                    a_i = aq_func(x_i.unsqueeze(0))
                    a.append(a_i.detach().cpu())
                a = torch.concat(a).unsqueeze(1)
                a_joint = aq_func(X_t).repeat(X_t.shape[0]).unsqueeze(1)  # Rearrange into m x 1

                eval_suggest['aq Value'] = a.numpy()
                
            eval_suggest['aq Value (joint)'] = a_joint.detach().cpu().numpy()
            eval_suggest['aq Method'] = [aq_str]*X_t.shape[0]

        # For multi-output evaluations, calculate pareto and hv considering objectives
        if o_dim > 1:
            if objective is None:
                o = f_t
                o_list = [f_t, f_t_train]
                if X_t_pending is not None:
                    o_list.append(f_t_pending)
            else:
                o_list = [o, o_train]
                if X_t_pending is not None:
                    o_list.append(o_pending)
            o_all = torch.concat(o_list, dim=0)

            hv = self.hypervolume(o_all)
            eval_suggest['Expected Hypervolume (joint)'] = hv
            pf = self.pareto(o_all)
            eval_suggest['Expected Pareto'] = pf[-o.shape[0]:]

        return eval_suggest

    def maximize(self,
                 optim_samples=1026,
                 optim_restarts=50,
                 fixed_var: dict[str: float | str] | None = None) -> tuple[pd.DataFrame, pd.DataFrame]:
        """
        Predicts the conditions which return the maximum response value within the parameter space.

        Args:
            optim_samples (int): The number of samples to be used for optimization. Default is ``1026``.
            optim_restarts (int): The number of restarts for the optimization process. Default is ``50``.
            fixed_var (dict(str:float), optional): Name of a variable and setting, over which the
                            suggestion should be fixed. Default values is ``None``
        Returns:
            tuple[pd.DataFrame, pd.DataFrame] = (X_suggest, eval_suggest)
                X_suggest (pd.DataFrame): Experiment matrix of real input variables,
                    selected by optimizer.
                y_suggest (pd.DataFrame): Mean results and prediction interval for
                    each suggested experiment.
        """
        
        X_suggest = pd.DataFrame()
        eval_suggest = pd.DataFrame()

        for target in self.target:
            X_suggest_i, eval_suggest_i = self.suggest(
                m_batch=1, acquisition=['Mean'], optim_samples=optim_samples, optim_restarts=optim_restarts,
                target=target, fixed_var=fixed_var)
            X_suggest = pd.concat([X_suggest, X_suggest_i], axis=0)
            eval_suggest = pd.concat([eval_suggest, eval_suggest_i], axis=0)
        
        return X_suggest, eval_suggest



"""Wrapper class for combining multiple objectives into a single objective."""

from .base import Objective
from .config import obj_class_dict

from torch import Tensor


class Objective_Sequence(Objective):
    """
    A composite objective that combines multiple acquisition
    objectives into a single objective, potentially single or multi-output

    Attributes:
        obj_list (list[Objective]): The list of acquisition objectives.

    Raises:
        TypeError: If the input is not a list of Objective objects.
    """

    def __init__(self,
                 obj_list: list[Objective]) -> None:
        if not isinstance(obj_list, list):
            raise TypeError('Objective_Sequence input must be a list of objectives')
        if not all(isinstance(obj, Objective) for obj in obj_list):
            raise TypeError('All elements in the list must be Objective objects')
        
        super().__init__(mo=obj_list[-1]._is_mo)

        self.obj_list = obj_list

        # None but the last objective can squeeze the multi-output dimension (if at all)
        for obj in obj_list[:-1]:
            obj._is_mo = True
    
    def __repr__(self):
        """String representation of object"""
        return f'{self.__class__.__name__} (obj_list={self.obj_list})'
    
    def forward(self,
                samples: Tensor,
                X: Tensor | None = None) -> Tensor:
        """
        Evaluate the objective function(s) on the candidate set samples, X
        """
        for obj in self.obj_list:
            samples = obj.forward(samples, X=X)

        return samples
    
    def save_state(self) -> dict:
        """Saves the objective(s) to a state dictionary"""
        obj_dict = {'name': self.__class__.__name__,
                    'obj_list': [obj.save_state() for obj in self.obj_list]}
        
        return obj_dict
    
    @classmethod
    def load_state(cls, obj_dict: dict):
        """Loads the objective(s) from a state dictionary"""
        new_obj_list = []
        
        for obj_dict_i in obj_dict['obj_list']:
            obj_class = obj_class_dict[obj_dict_i['name']]
            new_obj_list.append(obj_class.load_state(obj_dict_i))
 
        return cls(new_obj_list)



"""Surrogate model class definition"""

from obsidian.config import TORCH_DTYPE

from abc import ABC, abstractmethod

import numpy as np
import pandas as pd
import torch
import random


class SurrogateModel(ABC):
    """
    The model used for conducting optimization. Consumes data and produces a regressed representation
    of that system. Model can then be used to make predictions or evaluate uncertainty.

    Attributes:
        is_fit (bool): Flag for fitting (e.g. to prevent predicting from an unfit model).
        model_type (str): The type of the model.
        train_X (pd.DataFrame): The input data for the training data.
        train_Y (pd.Series): The target data for the training data.
        cat_dims (list): The categorical dimensions of the data.
        task_feature (str): The task feature of the data.
        X_order (list): The order of the columns in the input data.
        y_name (str): The name of the target column.
        seed (int): Randomization seed for stochastic surrogate models.
        verbose (bool): Flag for monitoring and debugging optimization
    """
    def __init__(self,
                 model_type: str = 'GP',
                 seed: int | None = None,
                 verbose: bool = False):
        
        # Set a flag for fitting (e.g. to prevent predicting from an unfit model)
        self.is_fit = False

        # Set up initial model state
        self.model_type = model_type
        self.train_X = None
        self.train_Y = None
        self.cat_dims = None
        self.task_feature = None
        self.X_order = None
        self.y_name = None

        # Handle randomization seed, considering all 3 sources (torch, random, numpy)
        self.seed = seed
        if self.seed is not None:
            torch.manual_seed(self.seed)
            torch.use_deterministic_algorithms(True)
            np.random.seed(self.seed)
            random.seed(self.seed)

        # Verbose output
        self.verbose = verbose

        return

    def _validate_data(self,
                       X: pd.DataFrame,
                       y: pd.Series | None = None):
        """
        Validate the data properties, e.g. column order
        
        Args:
            X (pd.DataFrame): Input parameters for the training data.
            y (pd.Series): Training data responses.
        
        Raises:
            ValueError: If the data properties do not match the training data.
        """

        if X.columns.tolist() != self.X_order:
            raise ValueError('X columns do not match training data')
        if y is not None:
            if y.name != self.y_name:
                raise ValueError('y column does not match training data')
        return
    
    def _prepare(self,
                 X: pd.DataFrame,
                 y: pd.Series | None = None) -> tuple:
        """
        Converts X, Y into appropriate Tensor dtypes and shapes for torch_model
        
        Args:
            X (pd.DataFrame): The input data as a pandas DataFrame.
            Y (pd.Series): The target data as a pandas Series.
        
        Returns:
            tuple: A tuple containing the converted input data (X_torch) and target data (Y_torch) as torch Tensors.
        """
        self._validate_data(X, y)
        X_torch = torch.tensor(X.to_numpy(), dtype=TORCH_DTYPE)
        if y is not None:
            y_torch = torch.tensor(y.to_numpy(), dtype=TORCH_DTYPE).unsqueeze(-1)
            return (X_torch, y_torch)
        else:
            return X_torch

    @abstractmethod
    def fit(self,
            X: pd.DataFrame,
            y: pd.Series):
        """Fit the surrogate model to data"""
        pass  # pragma: no cover

    @abstractmethod
    def predict(self,
                X: pd.DataFrame):
        """Predict outputs based on candidates X"""
        pass  # pragma: no cover

    @abstractmethod
    def score(self,
              X: pd.DataFrame,
              y: pd.Series):
        """Score the model based on the given test data"""
        pass  # pragma: no cover
    
    @abstractmethod
    def save_state(self):
        """Save the model to a state dictionary"""
        pass  # pragma: no cover
    
    @classmethod
    @abstractmethod
    def load_state(cls,
                   obj_dict: dict):
        """Load the model from a state dictionary"""
        pass  # pragma: no cover



"""Custom plots for SHAP analysis visualization"""

from .branding import obsidian_cm, obsidian_colors
from obsidian.exceptions import UnsupportedError

from shap.plots._partial_dependence import compute_bounds

import matplotlib.pyplot as plt
from matplotlib.axes import Axes
from matplotlib.figure import Figure

import numpy as np
import pandas as pd
from pandas.api.types import is_numeric_dtype

from typing import Callable


def one_shap_value(shap_value_new: np.ndarray,
                   expected_value: float,
                   X_names: list[str]) -> tuple[Figure, Figure]:
    """
    Visualize the shap values of one data point
    
    Args:
        shap_value_new (np.ndarray): The SHAP values of a single data point
            to be compared to a reference point.
        expected_value (float): The expected value at the reference point.
        X_names (list[str]): The names of the features.
        
    Returns:
        Figure: The bar plot of SHAP values for the single data point.
        Figure: The line plot of cumulative SHAP values for the data point in
            comparison to the reference point.
            
    """
    
    # First figure = Bar plot, just SHAP values of the new point
    pred_new = expected_value + np.sum(shap_value_new)
    
    fig_bar = plt.figure(figsize=(8, 6))
    plt.bar(range(len(shap_value_new)), shap_value_new, color=obsidian_colors.primary.teal)
    
    plt.xticks(range(len(shap_value_new)), X_names)
    plt.ylabel('SHAP Value')
    plt.xlabel('Feature')
    plt.title('SHAP Values for Single Datapoint')
    plt.axhline(0, color='grey', linestyle='--')
    plt.close(fig_bar)

    # Second figure = Line plot, cumulative SHAP values from new point to reference
    fig_line = plt.figure(figsize=(8, 6))
    
    cumulative_shap = np.cumsum(shap_value_new)
    y = cumulative_shap+expected_value
    n = len(cumulative_shap)
    
    plt.plot(range(n), y, color=obsidian_colors.primary.teal)
    plt.scatter(range(n), y, color=obsidian_colors.secondary.light_teal)
    
    plt.vlines(0, expected_value, y[0], colors='steelblue', linestyles='solid')
    plt.scatter(0, expected_value, color=obsidian_colors.accent.vista_blue)
    plt.scatter(n-1, y[-1], color=obsidian_colors.secondary.blue)
    
    plt.xticks(range(n), X_names)
    plt.xlabel('Feature')
    plt.ylabel('Cumulative SHAP Value')
    plt.title('Cumulative SHAP Values for Single Datapoint')
    plt.axhline(expected_value, color=obsidian_colors.accent.vista_blue,
                linestyle='--', label='Reference')
    plt.axhline(pred_new, color=obsidian_colors.secondary.blue,
                linestyle='--', label='New Data')
    plt.ylim([min(min(y), expected_value)*0.98, max(max(y), expected_value)*1.02])
    plt.legend()
    plt.close(fig_line)
    
    return fig_bar, fig_line


def partial_dependence(ind: int | tuple[int],
                       model: Callable,
                       data: pd.DataFrame,
                       ice_color_var: int | None = None,
                       xmin: str | tuple[float] | float = "percentile(0)",
                       xmax: str | tuple[float] | float = "percentile(100)",
                       npoints: int | None = None,
                       hist: bool = False,
                       ylabel: str | None = None,
                       ice: bool = True,
                       ace_opacity: float = 1,
                       pd_opacity: float = 1,
                       pd_linewidth: float = 2,
                       ace_linewidth: str | float = 'auto',
                       ax: Axes | None = None,
                       show: bool = True) -> Figure:
    """
    Calculates and plots the partial dependence of a feature or a pair of features on the model's output.

    This function is revised from the partial_dependence_plot function in shap package,
    in order to color the ICE curves by certain feature for checking interaction between features.
    Ref: https://github.com/shap/shap/blob/master/shap/plots/_partial_dependence.py

    Args:
        ind (int | tuple): The index or indices of the feature(s) to calculate
            the partial dependence for.
        model (Callable): The model used for prediction.
        data (pd.DataFrame): The input data used for prediction.
        ice_color_var (int, optional): The index of the feature used for coloring
            the ICE lines (for 1D partial dependence plot). Default is ``0``.
        xmin (str | tuple | float, optional): The minimum value(s) for the feature(s) range.
            Default is ``"percentile(0)"``.
        xmax (str | tuple | float): The maximum value(s) for the feature(s) range.
            Default is ``"percentile(100)"``.
        npoints (int, optional): The number of points to sample within the feature(s) range.
            By default, will use ``100`` points for 1D PDP and ``20`` points for 2D PDP.
        hist (bool, optional): Whether to plot the histogram of the feature(s). Default
            is ``False``.
        ylabel (str, optional): The label for the y-axis. Default is ``None``.
        ice (bool, optional): Whether to plot the Individual Conditional Expectation (ICE) lines.
            Default is ``True``.
        ace_opacity (float, optional): The opacity of the ACE lines. Default is ``1``.
        pd_opacity (float, optional): The opacity of the PDP line. Default is ``1``.
        pd_linewidth (float, optional): The linewidth of the PDP line. Default is ``2``.
        ace_linewidth (float | str, optional): The linewidth of the ACE lines. Default is ``'auto'``
            for automatic calculation.
        ax (Axes, optional): The matplotlib axis to plot on. By default will attach to Figure.gca().
        show (bool, optional): Whether to show the plot. Default is ``True``.

    Returns:
        tuple: A tuple containing the matplotlib figure and axis objects if `show` is False, otherwise None.
    """

    # Extract vals, names from data
    df_features = data
    feature_names = df_features.columns

    # 1D PDP
    if not isinstance(ind, tuple):
        # xv = values, independent variable
        # xs = scaled values
        xv = df_features.iloc[:, ind]
        df_test = df_features.copy()
        if not is_numeric_dtype(xv):
            xs = np.array(sorted(set(xv)))
            npoints = len(xs)
            xmin = xs[0]
            xmax = xs[-1]
        else:
            xmin, xmax = compute_bounds(xmin, xmax, xv)
            npoints = 100 if npoints is None else npoints
            xs = np.linspace(xmin, xmax, npoints).astype(xv.dtype)
        if ice:
            df_test = df_features.copy()
            ice_vals = []
            for i in range(npoints):
                df_test.iloc[:, ind] = xs[i]
                ice_vals.append(model(df_test))
            ice_vals = np.array(ice_vals)
        
        vals = []
        for i in range(npoints):
            df_test.iloc[:, ind] = xs[i]
            vals.append(model(df_test).mean())
        vals = np.array(vals)
        
        if ax is None:
            fig = plt.figure()
            ax1 = plt.gca()
        else:
            fig = plt.gcf()
            ax1 = plt.gca()

        ax2 = ax1.twinx()

        # Histogram
        if hist:
            ax2.hist(sorted(xv), 50, density=False, facecolor='black', alpha=0.1)

        # ICE line plot
        if ice:
            if ace_linewidth == "auto":
                ace_linewidth = min(1, 50/ice_vals.shape[1])
            if ice_color_var is None:
                ax1.plot(xs, ice_vals, color=obsidian_colors.secondary.light_teal,
                         linewidth=ace_linewidth, alpha=ace_opacity)
            else:
                if ice_color_var == ind:
                    raise UnsupportedError("Coloring by the feature(s) used in the PDP is not supported.")
                colormap = obsidian_cm.obsidian_viridis
                xc = df_features.iloc[:, ice_color_var]
                
                if not is_numeric_dtype(xc):
                    xc_cat_vals = sorted(set(xc))
                    colorbar_min = 0
                    colorbar_max = len(xc_cat_vals) - 1
                    color_vals = [xc_cat_vals.index(c)/colorbar_max for c in xc]
                else:
                    color_vals = xc
                    colorbar_min = color_vals.min()
                    colorbar_max = color_vals.max()
                    color_vals = ((color_vals - colorbar_min)/(colorbar_max-colorbar_min))
                    
                for i in range(ice_vals.shape[1]):
                    ax1.plot(xs, ice_vals[:, i], color=colormap(color_vals[i]),
                             linewidth=ace_linewidth, alpha=ace_opacity)
                cbar = plt.colorbar(plt.cm.ScalarMappable(cmap=colormap,
                                                          norm=plt.Normalize(
                                                              vmin=colorbar_min, vmax=colorbar_max)),
                                    ax=ax1)
                if not is_numeric_dtype(xc):
                    cbar.set_ticks(np.linspace(0, colorbar_max, len(xc_cat_vals)))
                    cbar.set_ticklabels(xc_cat_vals)
                cbar.set_label('Color by ' + feature_names[ice_color_var])

        ax1.plot(xs, vals, color="black", linewidth=pd_linewidth, alpha=pd_opacity)

        ax2.set_ylim(0, df_features.shape[0])
        ax1.set_xlabel(feature_names[ind], fontsize=13)
        if ylabel is None:
            if not ice:
                ylabel = "E[f(x) | " + str(feature_names[ind]) + "]"
            else:
                ylabel = "f(x) | " + str(feature_names[ind])

        ax1.set_ylabel(ylabel, fontsize=13)
        ax1.xaxis.set_ticks_position('bottom')
        ax1.yaxis.set_ticks_position('left')
        ax1.spines['right'].set_visible(False)
        ax1.spines['top'].set_visible(False)
        ax1.tick_params(labelsize=11)

        ax2.xaxis.set_ticks_position('bottom')
        ax2.yaxis.set_ticks_position('left')
        ax2.yaxis.set_ticks([])
        ax2.spines['right'].set_visible(False)
        ax2.spines['top'].set_visible(False)
        ax2.spines['left'].set_visible(False)
        ax2.spines['bottom'].set_visible(False)

        if show:
            plt.show()
        else:
            return fig, ax1

    # 2D PDP
    else:
        # xv = values, independent variable
        # xs = scaled values
        # x = scaled values, ordered, then cast to original type
        if ind[0] == ind[1]:
            raise UnsupportedError("The two features must be different for 2D PDP.")
        xv = []
        for i in ind:
            xv.append(df_features.iloc[:, i])

        xmin = list(xmin) if isinstance(xmin, tuple) else list([xmin, xmin])
        xmax = list(xmax) if isinstance(xmax, tuple) else list([xmax, xmax])
        xs = []
        x = []
        cat_list = []
        npoints = [20, 20] if npoints is None else [npoints, npoints]

        for i, xv_i in enumerate(xv):
            if not is_numeric_dtype(xv_i):
                cat_list.append(sorted(set(xv_i)))
                xmin[i] = 0
                xmax[i] = len(cat_list[i])
                xs.append(np.linspace(xmin[i], xmax[i]-1e-6, npoints[i]))
                x.append([cat_list[i][int(xi)] for xi in xs[i]])
                npoints[i] = len(xs[i])

            else:
                xmin[i], xmax[i] = compute_bounds(xmin[i], xmax[i], xv_i)
                xs.append(np.linspace(xmin[i], xmax[i], npoints[i]))
                x.append(np.linspace(xmin[i], xmax[i], npoints[i]).astype(xv_i.dtype))
                cat_list.append(None)

        df_test = df_features.copy()
        vals = []
        for i in range(npoints[0]):
            for j in range(npoints[1]):
                df_test.iloc[:, ind[0]] = x[0][i]
                df_test.iloc[:, ind[1]] = x[1][j]
                vals.append(model(df_test).mean())
        vals = np.array(vals).reshape(npoints[0], npoints[1])

        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')

        x0, x1 = np.meshgrid(xs[0], xs[1], indexing='ij')
        ax.plot_surface(x0, x1, vals, cmap=obsidian_cm.obsidian_viridis)

        ax.set_xlabel(feature_names[ind[0]], fontsize=13)
        if not is_numeric_dtype(xv[0]):
            ax.set_xticks(np.arange(len(cat_list[0])))
            ax.set_xticklabels(cat_list[0])
        
        ax.set_ylabel(feature_names[ind[1]], fontsize=13)
        if not is_numeric_dtype(xv[1]):
            ax.set_yticks(np.arange(len(cat_list[1])))
            ax.set_yticklabels(cat_list[1])
            
        ax.set_zlabel("E[f(x) | " + str(feature_names[ind[0]]) + ", " + str(feature_names[ind[1]]) + "]", fontsize=13)

        if show:
            plt.show()
        else:
            return fig, ax



import dash_bootstrap_components as dbc
from dash import dcc, html, Dash, dash_table, callback, Output, Input, State, ALL, MATCH
from dash.dash_table.Format import Format, Scheme
import dash_daq as daq
import pandas as pd

from obsidian.parameters import ParamSpace, Param_Categorical, Param_Ordinal, Param_Continuous
from obsidian.optimizer import BayesianOptimizer


def center(element):
    return html.Div(html.Div(element, style={'display': 'inline-block'}), style={'textAlign': 'center'})


def load_optimizer(config, opt_save):
    optimizer = BayesianOptimizer.load_state(opt_save)
    return optimizer


def load_Xspace(config, Xspace_save):
    X_space = ParamSpace.load_state(Xspace_save)
    return X_space


def add_tab(target, elements, id, label):
    target.children = list(target.children) if target.children else []
    tab = dbc.Tab(elements, label=label, id=id)
    target.children.append(tab)
    return target.children


def make_input(property_name, help_text, default_value=None, id=None, kwargs={}, required=True):
    components = [
        dbc.Label(property_name, style={'font-weight': 'bold', 'font-size': '0.8em'}),
        dbc.Input(value=default_value, id=f'input-[{property_name}]' if id is None else id,
                  debounce=True, required=required, **kwargs),
        html.Div(f'{help_text}', style={'font-size': '0.7em'}),
    ]
           
    return html.Div(components, className='mb-4')


def make_dropdown(property_name, help_text, options=[], id=None, kwargs={}):
    components = [
        dbc.Label(property_name, style={'font-weight': 'bold', 'font-size': '0.8em'}),
        dcc.Dropdown(options, id=f'input-[{property_name}]' if id is None else id, clearable=False, **kwargs),
        dbc.FormText(f'{help_text}', style={'font-size': '0.7em'}),
        ]

    return html.Div(components, className='mb-4')


def make_switch(property_name, help_text, id=None, kwargs={}):
    components = [
        dbc.Label(property_name, style={'font-weight': 'bold', 'font-size': '0.8em'}),
        html.Div(dbc.Switch(id=f'toggle-[{property_name}]' if id is None else id, value=True, **kwargs)),
        dbc.FormText(f'{help_text}', style={'font-size': '0.7em'}),
    ]

    return html.Div(components)


def make_slider(property_name, help_text, min, max, id=None, kwargs={}):
    components = [
        dbc.Label(property_name, style={'font-weight': 'bold', 'font-size': '0.8em'}),
        dcc.Slider(min, max, id=f'input-[{property_name}]' if id is None else id, **kwargs),
        dbc.FormText(f'{help_text}', style={'font-size': '0.7em'}),
    ]

    return html.Div(components)


def make_knob(property_name, help_text, min, max, id=None, kwargs={}):
    components = [
        dbc.Label(property_name, style={'font-weight': 'bold', 'font-size': '0.8em'}),
        daq.Knob(min=min, max=max, id=f'input-[{property_name}]' if id is None else id,  **kwargs),
        dbc.FormText(f'{help_text}', style={'font-size': '0.7em'}),
    ]

    return html.Div(components)


def make_table(df, fill_width=False):
    table = html.Div([dash_table.DataTable(data=df.to_dict('records'),
                                           columns=[{'id': c, 'name': c, 'type': 'numeric',
                                                     'format': {'specifier': '.5g'}} for c in df.columns],
                                           page_size=10, style_table={'overflowY': 'auto', 'overflowX': 'auto'},
                                           style_cell={'textAlign': 'center'}, style_header={'fontWeight': 'bold'},
                                           style_data_conditional=[{'if': {'row_index': 'odd'},
                                                                    'backgroundColor': 'rgb(220, 220, 220)'}],
                                           fill_width=fill_width,
                                           )], className='dbc')
    
    return table


def make_collapse(id, contents, label):
    components = [
        html.Div(dbc.Button(label, id=f'button-collapse-{id}', className='mb-3', color='primary', n_clicks=0),
                 style={'textAlign': 'center'}),
        dbc.Collapse(contents, id=f'collapse-{id}', is_open=False)
    ]
    return dbc.Card(dbc.CardBody(components))



"""Custom implementations of PyTorch surrogate models using BoTorch API"""

from .utils import fit_pytorch

from obsidian.config import TORCH_DTYPE

from botorch.models.model import FantasizeMixin
from botorch.models.ensemble import EnsembleModel, Model
from botorch.posteriors.ensemble import Posterior, EnsemblePosterior

import torch
import torch.nn as nn
from torch.nn import Module
from torch import Tensor

from typing import TypeVar
TFantasizeMixin = TypeVar("TFantasizeMixin", bound="FantasizeMixin")


class DNNPosterior(EnsemblePosterior):
    
    def __init__(self, values: Tensor):
        super().__init__(values)
         
    def quantile(self, value: Tensor) -> Tensor:
        """Quantile of the ensemble posterior"""
        return self.values.quantile(q=value.to(self.values), dim=-3, interpolation='linear')


class DNN(EnsembleModel, FantasizeMixin):
    def __init__(self,
                 train_X: Tensor,
                 train_Y: Tensor,
                 p_dropout: float = 0.2,
                 h_width: int = 16,
                 h_layers: int = 2,
                 num_outputs: int = 1):
        
        super().__init__()
        
        if h_layers < 1:
            raise ValueError("h_layers must be at least 1")
        if p_dropout < 0 or p_dropout > 1:
            raise ValueError("p_dropout must be in [0, 1]")
        
        self.register_buffer('train_X', train_X)
        self.register_buffer('train_Y', train_Y)
        self.register_buffer('p_dropout', torch.tensor(p_dropout, dtype=TORCH_DTYPE))
        self.register_buffer('h_width', torch.tensor(h_width, dtype=torch.int))
        self.register_buffer('h_layers', torch.tensor(h_layers, dtype=torch.int))
        self.register_buffer('num_outputs', torch.tensor(num_outputs, dtype=torch.int))

        self.input_layer = nn.Sequential(
            nn.Linear(train_X.shape[-1], h_width),
            nn.PReLU(),
            nn.Dropout(p=p_dropout)
        )

        self.middle_layers = nn.Sequential(
            *[nn.Sequential(
                nn.Linear(h_width, h_width),
                nn.PReLU(),
                nn.Dropout(p=p_dropout)
            ) for _ in range(h_layers)]
        )
        
        self.outer_layer = nn.Linear(h_width, num_outputs)
        self._num_outputs = num_outputs
        self.to(TORCH_DTYPE)
        
    def forward(self,
                x: Tensor) -> Tensor:
        x = self.input_layer(x)
        x = self.middle_layers(x)
        x = self.outer_layer(x)
        """Evaluate the forward pass of the model on inputs X"""
        return x

    def posterior(self,
                  X: Tensor,
                  n_sample: int = 512,
                  output_indices: list[int] = None,
                  observation_noise: bool | Tensor = False) -> Posterior:
        """Calculates the posterior distribution of the model at X"""
        if not output_indices:
            output_indices = list(range(self._num_outputs))
        elif not all(0 <= i < self._num_outputs for i in output_indices):
            raise ValueError("Invalid output index")
        
        if X.ndim == 2:
            X_sample = X.unsqueeze(0).repeat_interleave(n_sample, 0)
        else:  # Account for possible batch dimension
            X_sample = X.unsqueeze(1).repeat_interleave(n_sample, 1)
            
        self.train()
        y_out = self.forward(X_sample)
        self.eval()
        
        post = DNNPosterior(y_out[..., output_indices])
        
        return post
    
    @property
    def num_outputs(self) -> int:
        """Number of outputs of the model"""
        return self._num_outputs
    
    def transform_inputs(self,
                         X: Tensor,
                         input_transform: Module = None) -> Tensor:
        """
        Transform inputs.

        Args:
            X: A tensor of inputs
            input_transform: A Module that performs the input transformation.

        Returns:
            A tensor of transformed inputs
        """
        if input_transform is not None:
            input_transform.to(X)
            return input_transform(X)
        try:
            return self.input_transform(X)
        except AttributeError:
            return X

    def condition_on_observations(self,
                                  X: Tensor,
                                  Y: Tensor) -> TFantasizeMixin:
        """
        Condition the model to new observations, returning a fantasy model
        """

        X_c = torch.concat((self.train_X, X), axis=0)
        Y_c = torch.concat((self.train_Y, Y), axis=0)

        # Create a new model based on the current one
        fantasy = self.__class__(train_X=X_c, train_Y=Y_c,
                                 p_dropout=float(self.p_dropout),
                                 h_width=int(self.h_width), h_layers=int(self.h_layers),
                                 num_outputs=int(self.num_outputs))

        # Fit to the new data
        fit_pytorch(fantasy, X_c, Y_c)

        return fantasy
    
    def fantasize(self,
                  X: Tensor) -> Model:
        
        Y_f = self.forward(X).detach()
        fantasy = self.condition_on_observations(X, Y_f)

        return fantasy



"""BoTorch acquisition functions supported in obsidian"""

# Botorch acquisition functions supported in obsidian
from botorch.acquisition import qProbabilityOfImprovement, qUpperConfidenceBound, qSimpleRegret
from botorch.acquisition.logei import qLogExpectedImprovement, qLogNoisyExpectedImprovement
from botorch.acquisition.multi_objective.logei import qLogExpectedHypervolumeImprovement, qLogNoisyExpectedHypervolumeImprovement
from botorch.acquisition.multi_objective.parego import qLogNParEGO
from botorch.acquisition.active_learning import qNegIntegratedPosteriorVariance


__all__ = [
    "qProbabilityOfImprovement",
    "qUpperConfidenceBound",
    "qSimpleRegret",
    "qLogExpectedImprovement",
    "qLogNoisyExpectedImprovement",
    "qLogExpectedHypervolumeImprovement",
    "qLogNoisyExpectedHypervolumeImprovement",
    "qLogNParEGO",
    "qNegIntegratedPosteriorVariance",
]



from .utils import add_tab, make_input, make_dropdown, make_switch, make_slider, make_knob, make_table, make_collapse
from .utils import load_optimizer, center
import dash_bootstrap_components as dbc
from dash import dcc, html, Dash, dash_table, callback, Output, Input, State, ALL, MATCH

import pandas as pd
import base64
import io
from dash.exceptions import PreventUpdate

from obsidian.experiment import ExpDesigner


def setup_predict(app, app_tabs):
    # Parameter Sapce Table Review
    xspace_df_div = dbc.Container([
        dbc.Button('Refresh Parameter Space Table', id='button-config', n_clicks=0, size='lg'),
        html.Br(),
        html.Br(),
        dbc.Card([
            dbc.CardHeader('Parameter Space'),
            dbc.CardBody([
                html.Div(id='div-xspace_df', children=[], style={'overflow-x': 'scroll'})
                ]),
            ]),
        ],
        style={'textAlign': 'center'}
    )

    template_div = dbc.Container([
        dbc.Button('Generate Template', id='button-template', n_clicks=0, size='lg'),
        html.Br(),
        html.Br(),
        dbc.Card([
            dbc.CardHeader('Example Prediction Input'),
            dbc.CardBody([
                html.Div(id='div-template', children=[], style={'overflow-x': 'scroll'})
                ]),
            ]),
            ],
        style={'textAlign': 'center'}
    )
    # candidates store
    store_template = dcc.Store(id='store-template', data={})
    
    # template download
    template_downloader = html.Div(children=[dbc.Button('Download Template Candidates', id='button-download_template',
                                                        className='me-2', color='primary'),
                                             dcc.Download(id='downloader-template')],
                                   style={'textAlign': 'center', 'margin-top': '15px'})

    default_data = pd.DataFrame()
    
    # Data upload
    uploader_1 = dcc.Upload(id='uploader-X1',
                            children=html.Div(['Upload Data: Drag and Drop or ',
                                               html.A('Select Files')]),
                            style={
                              'width': '100%', 'height': '60px',
                              'lineHeight': '60px', 'borderWidth': '1px',
                              'borderStyle': 'dashed', 'borderRadius': '5px',
                              'textAlign': 'center', 'margin': '10px'
                              },
                            multiple=False,
                            filename='Example Data'
                            )
    
    # Data store
    storage_X1 = dcc.Store(id='store-X1', data=default_data.to_dict())
    
    # Data+Prediction preview
    preview_1 = html.Div(id='table-X1', children=dbc.Card(
        [dbc.CardHeader('Input Data and Predictions'),
         dbc.CardBody(make_table(default_data, fill_width=True), id='table-X1-body'),
         dbc.CardFooter(html.Div([
             html.I(id='info-data_1', className='bi bi-info-circle-fill me-2'),
             dbc.Tooltip('Input data must be a CSV file and must include column headers for the input\
                         parameters and response variable(s), with one row per observation. Download\
                         template data (left) for example.',
                         target='info-data_1', placement='top', style={'text-transform': 'none'}),
             dbc.FormText('Example Data', color='info', style={'font-size': '1em', 'font-style': 'italic'},
                          id='table-X1-footer')
             ],
             style={'textAlign': 'center'}))
         ]))

    row1 = dbc.Row([dbc.Col([xspace_df_div], width=4),
                    dbc.Col([template_div, template_downloader], width=8)], style={'margin-top': '15px'})
    row2 = dbc.Row([uploader_1, preview_1], style={'margin-top': '15px'})
    
    # Add all of these elements to the app
    elements = [html.Br(), store_template, row1, row2, storage_X1]
    add_tab(app_tabs, elements, 'tab-predict', 'Predict')
    setup_predict_callbacks(app)
    
    return


def setup_predict_callbacks(app):
    # Display the x_space as a dataframe
    @app.callback(
        Output('button-config', 'n_clicks'),
        Output('div-xspace_df', 'children'),
        Output('div-xspace_df', 'style'),
        Input('button-config', 'n_clicks'),
        State('store-Xspace', 'data'),
        prevent_initial_call=True
    )
    def config_tableView(clicked, Xspace_save):
        if not Xspace_save:
            return 0, None, {'overflow-x': 'scroll'}
        
        Xspace_list = []
        for param in Xspace_save.keys():
            state_dict = Xspace_save[param]['state']
            row = {'Parameter': state_dict['name']}
            row['Min'] = state_dict.get('min', None)
            row['Max'] = state_dict.get('max', None)
            row['Categories'] = str(state_dict.get('categories', None))
            Xspace_list.append(row)
        
        df_xspace = pd.DataFrame(Xspace_list)
        tables = [center(make_table(df_xspace))]
        return 0, tables, {'overflow-x': 'scroll'}
    
    # Generate New Data template according to x_space
    @app.callback(
        Output('button-template', 'n_clicks'),
        Output('div-template', 'children'),
        Output('store-template', 'data'),
        Input('button-template', 'n_clicks'),
        State('store-fit', 'data'),
        State('store-config', 'data'),
    )
    def config_InputTemplate(clicked, opt_save, config):
        if config is None:
            return 0, None, {}
        optimizer = load_optimizer(config, opt_save)
        designer = ExpDesigner(optimizer.X_space, seed=0)
        df_template = designer.initialize(3, 'LHS')
        tables = [center(make_table(df_template))]
        return 0, tables, df_template.to_dict()
    
    # Download Template
    @app.callback(
        Output('downloader-template', 'data'),
        Input('button-download_template', 'n_clicks'),
        State('store-template', 'data'),
        prevent_initial_call=True  # otherwise, the download CSV appears right away
    )
    def download_template(n_clicks, data):
        df = pd.DataFrame(data)
        return dcc.send_data_frame(df.to_csv, 'NewDataTemplate.csv', index=False)
    
    # Save the uploaded data into the data-store
    @app.callback(
        Output('store-X1', 'data'),
        Input('uploader-X1', 'contents'),
    )
    def save_X1(contents):
        if contents is None:
            raise PreventUpdate
        content_type, content_string = contents.split(',')
        decoded = base64.b64decode(content_string)
        df = pd.read_csv(io.StringIO(decoded.decode('utf-8')))
        return df.to_dict()
    
    # View the uploaded data in a Data Table
    @app.callback(
        Output('table-X1-body', 'children'),
        Output('table-X1-footer', 'children'),
        Input('store-X1', 'data'),
        State('uploader-X1', 'filename'),
        State('store-fit', 'data'),
        State('store-config', 'data'),
    )
    def preview_X1(data, filename, opt_save, config):
        df = pd.DataFrame(data)
        optimizer = load_optimizer(config, opt_save)
        preds = optimizer.predict(df)
        df_output = pd.concat([df.reset_index(drop=True), preds.reset_index(drop=True)], axis=1)
        return make_table(df_output, fill_width=True), filename
    
    return



"""Parameters: Define the classification of input features"""

from .targets import *
from .base import *
from .continuous import *
from .discrete import *
from .param_space import *
from .transforms import *
from .utils import *



"""
Simulation functions used for an optimization hackathon (OptiThon) in March 2024
"""

import numpy as np
import pandas as pd
from scipy.special import gamma


def Vm_func(X):
    # Maximum rate as a function of X
    rate = np.exp(-2*(X-0.75)**2)
    return rate


def Km_func(X):
    # Michaelis constant as a function of X
    rate = 2-np.exp(-1*(X-0.4)**2)
    return rate


def kI_func(X):
    # Inhibition constant as a function of X
    rate = 0.4/(1+np.exp(-1*(X-0.2)**2))
    return rate


def response_1(X):
    if X.shape[1] != 3:
        raise ValueError(f'Input to response_1 must have 3 columns, not {X.shape[1]}')
    X1, X2, X3 = X.T

    # Calculate kinetic parameters
    Vm = 100*Vm_func(X2)
    Km = Km_func(X3)
    kI = kI_func(X2)
    
    # Enzyme inhibition rate law
    rate = Vm*X1/(0.001+Km+X1+(X1/kI)**2)

    # Normalize
    rate = rate/10.5
    return rate


def response_2(X):
    if X.shape[1] != 2:
        raise ValueError(f'Input to response_2 must have 2 columns, not {X.shape[1]}')
    X1, X2 = X.T
    X2 = X2*1 + 2  # (ensure that there is at least one reactor)

    # Gamma function used in RTD calc
    gamma_value = gamma(X2)
    
    # RTD for a pulse into N CSTRs
    E_curve = (X1**(X2-1))*((2*X2)**X2)*np.exp(-2*X2*X1)/(gamma_value)
    
    # Normalize
    E_curve = E_curve/1.62
    
    return E_curve


def OT_simulator(X, addNoise=False):
    if isinstance(X, pd.DataFrame):
        X = X.to_numpy()
    if isinstance(X, np.ndarray) and X.ndim == 1:
        X = X[np.newaxis, :]

    if X.shape[1] != 6:
        raise ValueError(f'Input to simulate must have 6 columns, not {X.shape[1]}')
    
    # Break down the data into 3 parts for separate functions
    X123 = X[:, 0:3]
    X45 = X[:, 3:5]
    X6 = X[:, 5]
    
    # First three cols represented in an enzyme kinetics problem
    y1 = response_1(X123)
    
    # Next two cols represented in a flow problem
    y2 = response_2(X45)
    
    # Total performance is a weighted sum of two problems
    y = 0.6*y1 + 0.4*y2
    
    # Final column has a mean 0 effect, but determines the scale of noise
    # Note: The overall problem is slightly heteroskedastic
    if addNoise:
        sd = 0.01+0.02*X6
        rel_error = np.random.normal(loc=1, scale=sd, size=y.shape[0])
        y *= rel_error
    
    return y



"""Campaign class definition"""

from obsidian.parameters import ParamSpace, Target
from obsidian.optimizer import Optimizer, BayesianOptimizer
from obsidian.experiment import ExpDesigner
from obsidian.objectives import Objective, Objective_Sequence, obj_class_dict
from obsidian.constraints import Output_Constraint, const_class_dict
from obsidian.exceptions import IncompatibleObjectiveError
from obsidian.utils import tensordict_to_dict
import obsidian

import pandas as pd
import torch
import warnings


class Campaign():
    """
    Base class for tracking optimization progress and other metrics
    over multiple iterations.

    Attributes:
        X_space (ParamSpace): The parameter space for the campaign.
        data (pd.DataFrame): The data collected during the campaign.
        optimizer (Optimizer): The optimizer used for optimization.
        designer (ExpDesigner): The experimental designer used for experiment design.
        iter (int): The current iteration number.
        seed (int): The seed for random number generation.
    
    Properties:
        m_exp (int): The number of observations in campaign.data
        y (pd.Series): The response data in campaign.data
        y_names (list): The names of the response data columns
        f (pd.Series): The transformed response data
        o (pd.Series): The objective function evaluated on f
        o_names (list): The names of the objective function columns
        X (pd.DataFrame): The input features of campaign.data
        response_max (float | pd.Series): The maximum for each response
        target (Target | list[Target]): The target(s) for optimization.
        objective (Objective, optional): The objective of the optimization campaign
    
    """
    
    def __init__(self,
                 X_space: ParamSpace,
                 target: Target | list[Target],
                 constraints: Output_Constraint | list[Output_Constraint] | None = None,
                 optimizer: Optimizer | None = None,
                 designer: ExpDesigner | None = None,
                 objective: Objective | None = None,
                 seed: int | None = None):
        
        self.set_X_space(X_space)
        self.data = pd.DataFrame()
        
        optimizer = BayesianOptimizer(X_space, seed=seed) if optimizer is None else optimizer
        self.set_optimizer(optimizer)

        designer = ExpDesigner(X_space, seed=seed) if designer is None else designer
        self.set_designer(designer)
        
        self.set_target(target)
        self.set_objective(objective)
        
        self.output_constraints = None
        self.constrain_outputs(constraints)
        
        # Non-object attributes
        self.iter = 0
        self.seed = seed
        self.version = obsidian.__version__

    def add_data(self, df: pd.DataFrame):
        """
        Adds data to the campaign.

        Args:
            Z_i (pd.DataFrame): The data to be added to the campaign.
        
        Raises:
            KeyError: If all X_names are not in the dataset
            KeyError: If all y_names are not in the dataset
        """
        
        if not all(name in df.columns for name in self.X_space.X_names):
            raise KeyError('Input dataset does not contain all of the required parameter names')
        if not all(name in df.columns for name in self.y_names):
            raise KeyError('Input dataset does not contain all of the required response target names')
        
        new_data = df

        if 'Iteration' not in new_data.columns:
            new_data['Iteration'] = self.iter
        else:
            self.iter = int(new_data['Iteration'].max())

        self.iter += 1
        self.data = pd.concat([self.data, new_data], axis=0, ignore_index=True)
        self.data.index.name = 'Observation ID'
        self.data.index = self.data.index.astype('int')
        
        if self.optimizer.is_fit:
            self._analyze()

    def clear_data(self):
        """Clears campaign data"""
        self.data = pd.DataFrame()
        self.iter = 0

    @property
    def X_space(self) -> ParamSpace:
        """Campaign ParamSpace"""
        return self._X_space
    
    def set_X_space(self, X_space: ParamSpace):
        """Sets the campaign ParamSpace"""
        self._X_space = X_space

    @property
    def optimizer(self) -> Optimizer:
        """Campaign Optimizer"""
        return self._optimizer
    
    def set_optimizer(self, optimizer: Optimizer):
        """Sets the campaign optimizer"""
        self._optimizer = optimizer
        
    @property
    def designer(self) -> ExpDesigner:
        """Campaign Experimental Designer"""
        return self._designer
    
    def set_designer(self, designer: ExpDesigner):
        """Sets the campaign experiment designer"""
        self._designer = designer

    @property
    def objective(self) -> Objective | None:
        """Campaign Objective function"""
        return self._objective
    
    def _eval_objective(self):
        """Evaluates objective and appends it to campaign data"""
        df_o = self.o
        for col in df_o.columns:
            self.data[col] = df_o[col].values
            self.o_names = [col for col in self.data.columns if 'Objective' in col]
    
    def set_objective(self, objective: Objective | None):
        """(Re)sets the campaign objective function"""
        self._objective = objective
        if not self.data.empty:
            # Remove previous objective evaluations
            self.data = self.data.drop(
                columns=[col for col in self.data.columns if 'Objective' in col]
                )
            if self.optimizer.is_fit:
                self._analyze()
                
    def clear_objective(self):
        """Clears the campaign objective function"""
        self._objective = None
                
    @property
    def target(self):
        """Campaign experimental target(s)"""
        return self._target

    def set_target(self,
                   target: Target | list[Target]):
        """
        Sets the experimental target context for the campaign.

        Args:
            target (Target | list[Target] | None): The target or list of targets to set.
            
        """
        if isinstance(target, Target):
            self._target = [target]
        else:
            self._target = target

        self.y_names = [t.name for t in self._target]
        self.n_response = len(self.y_names)

    @property
    def _is_mo(self) -> bool:
        """
        Boolean flag for multi-output
        """
        if self.objective:
            return self.objective._is_mo
        else:
            return self.n_response > 1
    
    @property
    def m_exp(self) -> int:
        """
        Number of observations in training data
        """
        return self.data.shape[0]

    @property
    def y(self) -> pd.Series | pd.DataFrame:
        """
        Experimental response data

        """
        if not self.data.empty:
            return self.data[self.y_names]
        else:
            return None

    @property
    def response_max(self) -> float | pd.Series:
        """
        Maximum response data in training set
        """
        return self.y.max()

    @property
    def f(self) -> pd.Series | pd.DataFrame:
        """
        Experimental response data, in transformed space
        """
        f = pd.concat([t.transform_f(self.y[t.name]) for t in self.target], axis=1)
        return f

    @property
    def o(self) -> pd.Series | pd.DataFrame:
        """
        Objective function evaluated on f
        """
        if self.objective:
            try:
                x = self.X_space.encode(self.X).values
                o = self.objective(torch.tensor(self.f.values).unsqueeze(0),
                                   X=torch.tensor(x)).squeeze(0)
                if o.ndim < 2:
                    o = o.unsqueeze(1)  # Rearrange into m x o
                return pd.DataFrame(o.detach().cpu().numpy(),
                                    columns=[f'Objective {o_i+1}' for o_i in range(o.shape[1])])
            except Exception:
                raise IncompatibleObjectiveError('Objective(s) did not successfully execute on sample')
        else:
            return None
    
    @property
    def out(self) -> pd.Series | pd.DataFrame:
        """
        Returns the objective function as appropriate, else the response data
        """
        if self.objective and self.optimizer.is_fit:
            return self.o
        else:
            return self.y
    
    @property
    def X_best(self) -> pd.DataFrame:
        """
        Best performing X values
        """
        best_idx = self.out.idxmax().values
        
        X_best = self.X.iloc[best_idx, :]
        if isinstance(X_best, pd.Series):
            X_best = X_best .to_frame().T
        
        return X_best
    
    @property
    def X(self) -> pd.DataFrame:
        """
        Feature columns of the training data
        """
        return self.data[list(self.X_space.X_names)]
            
    def __repr__(self):
        """String representation of object"""
        return f"obsidian Campaign for {getattr(self,'y_names', None)}; {getattr(self,'m_exp', 0)} observations"

    def initialize(self, **design_kwargs):
        """
        Maps ExpDesigner.initialize method
        """
        return self.designer.initialize(**design_kwargs)
    
    def fit(self):
        """
        Maps Optimizer.fit method

        Raises:
            ValueError: If no data has been registered to the campaign
        """

        if self.m_exp <= 0:
            raise ValueError('Must register data before fitting')

        self.optimizer.fit(self.data, target=self.target)

    def suggest(self, **optim_kwargs):
        """
        Maps Optimizer.suggest method
        """
        if self.optimizer.is_fit:
            try:
                # In case X_space has changed, re-set the optimizer X_space
                self.optimizer.set_X_space(self.X_space)
                X, eval = self.optimizer.suggest(objective=self.objective,
                                                 out_constraints=self.output_constraints,
                                                 **optim_kwargs)
                return (X, eval)
            except Exception:
                warnings.warn('Optimization failed')
                return None
        else:
            warnings.warn('Optimizer is not fit to data. Suggesting initial experiments.', UserWarning)
            X0 = self.initialize()
            return X0

    def evaluate(self, X_suggest: pd.DataFrame):
        """
        Maps Optimizer.evaluate method
        """
        return self.optimizer.evaluate(X_suggest, objective=self.objective)

    def _profile_hv(self):
        """
        Calculate and assign the hypervolume values to each iteration in the data.

        Returns:
            None
        """
        iters = self.data['Iteration'].unique()
        hv = {}
        
        for i in iters:
            iter_index = self.data.query(f'Iteration <= {i}').index
            out_iter = self.out.loc[iter_index, :]
            out_iter = torch.tensor(out_iter.values)
            hv[i] = self.optimizer.hypervolume(out_iter)
        
        self.data['Hypervolume (iter)'] = self.data.apply(lambda x: hv[x['Iteration']], axis=1)
        self.data['Pareto Front'] = self.optimizer.pareto(torch.tensor(self.out.values))
        
        return

    def _profile_max(self):
        """
        Calculate the maximum values achieved for targets at each iteration

        Returns:
            None
        """
        
        # Remove previous max-profiling
        self.data = self.data.drop(
            columns=[col for col in self.data.columns if '(max) (iter)' in col]
        )
        
        for out in self.out.columns:
            self.data[out+' (max) (iter)'] = self.data.apply(
                lambda x: self.data.query(f'Iteration<={x["Iteration"]}')[out].max(), axis=1
            )

        return
    
    def _analyze(self):
        """
        Analyzes the campaign data for practical optimization performance metrics

        Returns:
            None
        """
        if self.objective:
            self._eval_objective()
        self._profile_max()
        if self._is_mo:
            self._profile_hv()
        else:
            # Remove previous HV-profiling
            self.data = self.data.drop(
                columns=[col for col in self.data.columns
                         if 'Hypervolume' in col or 'Pareto' in col]
            )
            
        return

    def constrain_outputs(self,
                          constraints: Output_Constraint | list[Output_Constraint] | None) -> None:
        """
        Sets optional output constraints for the campaign.
        """
        if constraints is not None:
            if isinstance(constraints, Output_Constraint):
                constraints = [constraints]
            self.output_constraints = constraints

        return
    
    def clear_output_constraints(self):
        """Clears output constraints"""
        self.output_constraints = None

    def save_state(self) -> dict:
        """
        Saves the state of the Campaign object as a dictionary.

        Returns:
            dict: A dictionary containing the saved state of the Campaign object.
        """
        obj_dict = {}
        obj_dict['X_space'] = self.X_space.save_state()
        obj_dict['optimizer'] = self.optimizer.save_state()
        obj_dict['data'] = self.data.to_dict()
        obj_dict['target'] = [t.save_state() for t in self.target]
        if self.objective:
            obj_dict['objective'] = self.objective.save_state()
        obj_dict['seed'] = self.seed

        if getattr(self, 'output_constraints', None):
            obj_dict['output_constraints'] = [{'class': const.__class__.__name__,
                                               'state': tensordict_to_dict(const.state_dict())}
                                              for const in self.output_constraints]

        return obj_dict
    
    @classmethod
    def load_state(cls,
                   obj_dict: dict):
        """
        Loads the state of the campaign from a dictionary.

        Args:
            cls (Campaign): The class object.
            obj_dict (dict): A dictionary containing the campaign state.

        Returns:
            Campaign: A new campaign object with the loaded state.
        """
        
        if 'objective' in obj_dict:
            if obj_dict['objective']['name'] == 'Objective_Sequence':
                new_objective = Objective_Sequence.load_state(obj_dict['objective'])
            else:
                obj_class = obj_class_dict[obj_dict['objective']['name']]
                new_objective = obj_class.load_state(obj_dict['objective'])
        else:
            new_objective = None
        
        new_campaign = cls(X_space=ParamSpace.load_state(obj_dict['X_space']),
                           target=[Target.load_state(t_dict) for t_dict in obj_dict['target']],
                           optimizer=BayesianOptimizer.load_state(obj_dict['optimizer']),
                           objective=new_objective,
                           seed=obj_dict['seed'])
        new_campaign.data = pd.DataFrame(obj_dict['data'])
        new_campaign.data.index = new_campaign.data.index.astype('int')
        
        new_campaign.iter = new_campaign.data['Iteration'].astype('int').max()

        if 'output_constraints' in obj_dict:
            for const_dict in obj_dict['output_constraints']:
                const = const_class_dict[const_dict['class']](new_campaign.target, **const_dict['state'])
                new_campaign.constrain_outputs(const)

        return new_campaign



"""Transformation functions to normalize output responses"""

from torch import logit, sigmoid
from abc import ABC, abstractmethod
import warnings
from torch import Tensor

from obsidian.exceptions import UnfitError

# Method name pointers
f_transform_dict = {'Standard': lambda: Standard_Scaler(),
                    'Identity': lambda: Identity_Scaler(),
                    'Logit_MinMax': lambda: Logit_Scaler(),
                    'Logit_Percentage': lambda: Logit_Scaler(range_response=100, override_fit=True),
                    }


class Target_Transform(ABC):
    """
    Base class for obsidian Target transforms
    """
    def __init__(self):
        self.params = {}

    def _validate_fit(self):
        """
        Validates if all parameters have been fit before transforming.

        Raises:
            UnfitError: If any parameter value is None, indicating that the parameters have not been fit.
        """
        if not all([(v is not None) for v in self.params.values()]):
            raise UnfitError('Params must be fit before transforming.')

    @abstractmethod
    def forward(self,
                X: Tensor,
                fit: bool = False):
        """Evaluate the forward transformation on input data X"""
        pass  # pragma: no cover

    @abstractmethod
    def inverse(self,
                X: Tensor):
        """Inverse transform the transformed data X_t"""
        pass  # pragma: no cover

    def __call__(self,
                 X: Tensor,
                 fit: bool = False):
        """Shortcut to forward method"""
        return self.forward(X, fit)
        

class Identity_Scaler(Target_Transform):
    """
    Dummy scaler class which simply returns the input
    """
    def forward(self,
                X: Tensor,
                fit: bool = False):
        """Evaluate the forward transformation on input data X"""
        return X
    
    def inverse(self,
                X: Tensor):
        """Inverse transform the transformed data X_t"""
        return X


class Standard_Scaler(Target_Transform):
    """
    Scaler which normalizes based on zero mean and unit st-dev
    """
    def __init__(self):
        self.params = {'mu': None, 'sd': None}
    
    def forward(self,
                X: Tensor,
                fit: bool = False):
        """Evaluate the forward transformation on input data X"""
        if fit:
            X_v = X[~X.isnan()]
            self.params = {'mu': X_v.mean(), 'sd': X_v.std()}
        else:
            self._validate_fit()
        return (X-self.params['mu'])/self.params['sd']
    
    def inverse(self, X):
        """Inverse transform the transformed data X_t"""
        self._validate_fit()
        return X*self.params['sd']+self.params['mu']


class Logit_Scaler(Target_Transform):
    """
    Scaler which normalizes based on a logit transform
    Can be fit to select an appropriate range for the logit
    """
    def __init__(self,
                 range_response: int | float = 1,
                 loc: int | float = 0,
                 override_fit: bool = False,
                 standardize: bool = True):
        self.params = {'scale': 1/range_response, 'loc': loc, 'mu': None, 'sd': None}
        # Override "fitting" when valid ranges are provided during init
        self.override_fit = override_fit
        self.standardize = standardize

    def _fit_minmax(self,
                    X: Tensor):
        """Fits the min-max scale of the logit transform"""
        # Scale X into a range from 0-1 with buffer/2 on either side
        self.override_fit = False
        range_response = X.max()-X.min()
        buffer = 0.2
        self.params['scale'] = (1-buffer)/range_response
        self.params['loc'] = X.min() - (buffer/2)*(1/self.params['scale'])

    def forward(self,
                X: Tensor,
                fit: bool = False):
        """Evaluate the forward transformation on input data X"""
        # If fit is not called, and the range is valid, transform
        # If the range is invalid, fit the range first and warn
        if not fit or self.override_fit:
            X_s = self.params['scale']*(X - self.params['loc'])
            valid_range = (X_s >= 0).all() and (X_s <= 1).all()
            if not valid_range:
                warnings.warn('Invalid range provided for logit scaler, proceeding with min-max fit', UserWarning)
                self._fit_minmax(X)
                return self.forward(X)
        else:
            X_v = X[~X.isnan()]
            self._fit_minmax(X_v)
            X_s = self.params['scale']*(X - self.params['loc'])
        X_st = logit(X_s)
        if self.standardize:
            self.params.update({'mu': X_st.mean(), 'sd': X_st.std()})
            return (X_st-self.params['mu'])/self.params['sd']
        else:
            return X_st
    
    def inverse(self,
                X: Tensor):
        """Inverse transform the transformed data X_t"""
        if self.standardize:
            self._validate_fit()
            X = X*self.params['sd']+self.params['mu']
        return (1/self.params['scale'])*sigmoid(X)+self.params['loc']



"""Constraints on the output responses of a model"""

from .base import Constraint

from obsidian.parameters import Target
from obsidian.utils import unscale_samples
from obsidian.config import TORCH_DTYPE

import torch
from torch import Tensor
from typing import Callable


class Output_Constraint(Constraint):
    """
    Output constraint for a given set of targets.

    Must return a callable function that computes feasibility, where
        negative values imply feasible space.

    Note: Saving and loading input constraints is managed by Campaign
    """
    def __init__(self,
                 target: Target | list[Target]):
        super().__init__()
        self.target = self._validate_target(target)

    def _validate_target(self, target: Target | list[Target]):

        if not isinstance(target, (Target, list)):
            raise TypeError('Target must be a Target object or a list of Target objects')
        if isinstance(target, list):
            for t in target:
                if not isinstance(t, Target):
                    raise TypeError('Target must be a Target object or a list of Target objects')
        if isinstance(target, Target):
            target = [target]
        
        return target


class Blank_Constraint(Output_Constraint):
    """
    Dummy constraint function that proposes all samples as feasible.
    """
    def __init__(self,
                 target: Target | list[Target]):
        super().__init__(target)

    def forward(self,
                scale: bool = True) -> Callable:
        def constraint(samples: Tensor) -> Tensor:
            if scale:
                samples = unscale_samples(samples, self.target)
            feasibility = -1*torch.ones(size=samples.shape).max(dim=-1).values
            return feasibility
        return constraint
    
    def __repr__(self):
        """String representation of object"""
        return f'{self.__class__.__name__}'
    
    
class L1_Constraint(Output_Constraint):
    """
    Calculates the L1 (absolute-value penalized) constraint
    """
    def __init__(self,
                 target: Target | list[Target],
                 offset: int | float = 1):
        super().__init__(target)
        self.register_buffer('offset', torch.tensor(offset, dtype=TORCH_DTYPE))

    def forward(self,
                scale: bool = True) -> Callable:
        def constraint(samples: Tensor) -> Tensor:
            if scale:
                samples = unscale_samples(samples, self.target)
            feasibility = (samples.sum(dim=-1) - self.offset)
            return feasibility
        return constraint
    
    def __repr__(self):
        """String representation of object"""
        return f'{self.__class__.__name__}(offset={self.offset})'



"""Optimizer class definition"""

from obsidian.parameters import ParamSpace, Target, Param_Observational
from obsidian.exceptions import UnsupportedError

from botorch.utils.multi_objective.box_decompositions.dominated import DominatedPartitioning
from botorch.utils.multi_objective.pareto import is_non_dominated
from abc import ABC, abstractmethod

import pandas as pd
import numpy as np
import torch
import random
from torch import Tensor


class Optimizer(ABC):
    """
    Base class for obsidian optimizer, which fits a surrogate model to data and suggests
    optimal experiments

    Attributes:
        X_space (ParamSpace): obsidian ParamSpace object representing the allowable space for optimization.
        seed (int | None): Randomization seed for the optimizer and stochastic surrogate models.
        verbose (int): Flag for monitoring and debugging optimization output.

    Raises:
        ValueError: If verbose is not set to 0, 1, 2, or 3.
        TypeError: If X_space is not an obsidian ParamSpace
    """

    def __init__(self,
                 X_space: ParamSpace,
                 seed: int | None = None,
                 verbose: int = 1):
        
        # Verbose selection
        if verbose not in [0, 1, 2, 3]:
            raise ValueError('Verbose option must be 0 (no output), 1 (summary output), \
                             2 (detailed output), or 3 (debugging)')
        self.verbose = verbose

        # Handle randomization seed, considering all 3 sources (torch, random, numpy)
        self.seed = seed
        if self.seed is not None:
            torch.manual_seed(self.seed)
            torch.use_deterministic_algorithms(True)
            np.random.seed(self.seed)
            random.seed(self.seed)

        # Store the parameter space which contains useful reference properties
        if not isinstance(X_space, ParamSpace):
            raise TypeError('X_space must be an obsidian ParamSpace object')
        self.set_X_space(X_space)

    @property
    def X_space(self):
        """
        ParamSpace: The parameter space defining the search space for the optimization.
        """
        return self._X_space
    
    def set_X_space(self, X_space: ParamSpace):
        self._X_space = X_space
        return

    def _fixed_features(self,
                        fixed_var: dict | None = None) -> list:
        """
        Returns a list of fixed features for optimization.

        Args:
            fixed_var (dict): A dictionary of fixed variables and their corresponding settings.
                The dictionary should be in the format: {param_name: param_setting}.

        Returns:
            list: A list of fixed features for optimization.

        Raises:
            TypeError: If fixed_var is not a dictionary.
            NameError: If a parameter in fixed_var is not found in X_space.
        """

        # Set static optimization features (used for fitting, but set to a fixed number during optimization)
        # e.g. Optimizing for max yield at a fixed end-time in a reaction

        fixed_var = {} if fixed_var is None else fixed_var

        fixed_features_list = []

        # Add any observational variables to fixed_var
        if self.X_space.X_obs:
            fixed_obs = {p.name: p.design_point for p in self.X_space if isinstance(p, Param_Observational)}
            fixed_var.update(fixed_obs)

        # Store each validate setting list in a dataframe, then merge by cartesian product to get all valid combos
        df_list = []

        if fixed_var:
            if not isinstance(fixed_var, dict):
                raise TypeError('Fixed variables must be provided as a dict: {param_name:param_setting}')
            for var, level in fixed_var.items():
                if var not in self.X_space.X_names:
                    raise NameError(f'Profile variable {var} not found in {self.X_space.X_names}')
                param_i = next(param for param in self.X_space if param.name == var)
                param_i._validate_value(level)
                df_i = pd.DataFrame({var: [level]})
                df_list.append(df_i)

        # If we have discrete parameters, we need to limit the acquisition function search as such
        # First, get the cartesian product of all of the categorical/ordinal combos
        for x in self.X_space.X_discrete:
            if x.name not in fixed_var.keys():  # Fixed_var should take precedent and lock out other combinations
                df_i = pd.DataFrame({x.name: x.search_categories})
                df_list.append(df_i)
        
        # Merge by cross
        if df_list:
            df_all = df_list[0]
            for df_i in df_list[1:]:
                df_all = pd.merge(df_all, df_i, how='cross')
            encoded_cross = self.X_space.encode(df_all)

            # Fixed feature list requires names to be the indeces of our columns
            X_t_names = self.X_t_train.columns
            encoded_cross = encoded_cross.rename(columns={col: X_t_names.get_loc(col) for col in encoded_cross.columns})
            fixed_features_list += encoded_cross.to_dict(orient='records')

        return fixed_features_list

    def hypervolume(self,
                    f: Tensor,
                    ref_point: list | None = None,
                    weights: list | None = None,
                    ) -> float:
        """
        Calculates the hypervolume of the given data points.

        Args:
            f (Tensor): The data points to calculate the hypervolume for.
            ref_point (list, optional): The reference point for the hypervolume calculation. Defaults to ``None``.
            weights (list, optional): The weights to apply to each objective. Defaults to ``None``.

        Returns:
            float: The hypervolume value.

        Raises:
            UnsupportedError: If the number of objectives is less than or equal to 1.
        """

        if f.shape[1] <= 1:
            raise UnsupportedError('Cannot calculate hypervolume for single objective')
        if f.ndim != 2:
            raise ValueError('Hypervolume calculation only supported for 2D tensor')

        if ref_point is None:
            ref_point = f.min(dim=0).values
        else:
            if len(ref_point) != f.shape[1]:
                raise ValueError('Reference point must have the same number of objectives as the data')
            # Need to transform ref_point if provided
            ref_point = torch.tensor(ref_point)

        if weights is None:
            weights = torch.ones(size=(f.shape[1],))
        else:
            if len(weights) != f.shape[1]:
                raise ValueError('Weights must have the same number of objectives as the data')
            weights = torch.tensor(weights)

        f_weighted = f * weights
        bd = DominatedPartitioning(ref_point=ref_point, Y=f_weighted)
        hv = bd.compute_hypervolume().item()

        return hv

    def pareto(self,
               f: Tensor) -> list[bool]:
        """
        Determines the Pareto dominance of a given set of solutions.

        Args:
            f (Tensor): The input series containing the solutions.

        Returns:
            list[bool]: A list of boolean values indicating whether each solution is Pareto optimal or not.
        """
        
        if f.shape[1] <= 1:
            raise UnsupportedError('Cannot calculate pareto front for single objective')
        if f.ndim != 2:
            raise ValueError('Pareto front calculation only supported for 2D tensor')

        return is_non_dominated(f).tolist()
    
    def pf_distance(self,
                    y: pd.Series) -> Tensor:
        """
        Calculates the pairwise distance between the given input `y` and the Pareto front.

        Args:
            y (pd.Series): The input data.

        Returns:
            Tensor: The minimum distance between `y` and the Pareto front.
        """
        f = torch.tensor(pd.concat([t.transform_f(y[t.name]) for t in self.target], axis=1).to_numpy())
        pf = self.pareto(torch.tensor(self.f_train.values))
        f_pareto = torch.tensor(self.f_train.values)[pf]
        
        f = f.unsqueeze(1)  # Insert "to_xdim" before "from_xdim"
        f_pareto = f_pareto.unsqueeze(0)  # Insert "from_xdim" before "to_xdim"
        
        pairwise_distance = torch.norm(f - f_pareto, p=2, dim=-1)
        min_distance = pairwise_distance.min(dim=1).values
        
        return min_distance

    @abstractmethod
    def fit(self,
            Z: pd.DataFrame,
            target: Target | list[Target]):
        """Fit the optimizer's surrogate models to data"""
        pass  # pragma: no cover

    @abstractmethod
    def predict(self,
                X: pd.DataFrame,
                return_f_inv: bool = True,
                PI_range: float = 0.7):
        """Predict the optimizer's target(s) at the candidate set X"""
        pass  # pragma: no cover
    
    @abstractmethod
    def suggest(self):
        """Suggest the next optimal experiment(s)"""
        pass  # pragma: no cover
    
    @abstractmethod
    def maximize(self):
        """Maximize the optimizer's target(s)"""
        pass  # pragma: no cover

    @abstractmethod
    def save_state(self):
        """Save the optimizer to a state dictionary"""
        pass  # pragma: no cover
    
    @classmethod
    @abstractmethod
    def load_state(cls,
                   obj_dict: dict):
        """Load the optimizer from a state dictionary"""
        pass  # pragma: no cover



"""Method pointers and config for surrogate models"""

from .custom_GP import DKLGP, FlatGP, PriorGP
from .custom_torch import DNN

from botorch.models import SingleTaskGP, MixedSingleTaskGP, MultiTaskGP

# Surrogate model method pointer
model_class_dict = {'GP': SingleTaskGP,
                    'MixedGP': MixedSingleTaskGP,
                    'DKL': DKLGP,
                    'GPflat': FlatGP,
                    'GPprior': PriorGP,
                    'MTGP': MultiTaskGP,
                    'DNN': DNN}



"""Custom obsidian exceptions for improved error handling"""


class IncompatibleObjectiveError(Exception):
    """Exception that gets raised when the objective(s) cannot be successfully called as specified"""
    pass


class SurrogateFitError(Exception):
    """Exception that gets raised when the surrogate model fails to fit"""
    pass


class UnsupportedError(Exception):
    """Exception that gets raised when an optimization is requested on an unsupported feature"""
    pass


class UnfitError(Exception):
    """Exception that gets raised when an action is called before a model or transform has been fit"""
    pass


class DataWarning(UserWarning):
    """Warning that gets raised if there is an issue with input data"""
    pass


class OptimizerWarning(UserWarning):
    """Warning that gets raised if there is an issue with optimization configuration"""
    pass



"""Method pointers and config for objective functions"""

from .custom import (
    Identity_Objective,
    Feature_Objective,
    Utopian_Distance,
    Bounded_Target,
    Index_Objective,
    Product_Objective,
    Divide_Objective
)

from .scalarize import (
    Scalar_WeightedSum,
    Scalar_WeightedNorm,
    Scalar_Chebyshev,
)

obj_class_dict = {'Identity_Objective': Identity_Objective,
                  'Feature_Objective': Feature_Objective,
                  'Utopian_Distance': Utopian_Distance,
                  'Bounded_Target': Bounded_Target,
                  'Scalar_WeightedSum': Scalar_WeightedSum,
                  'Scalar_WeightedNorm': Scalar_WeightedNorm,
                  'Scalar_Chebyshev': Scalar_Chebyshev,
                  'Index_Objective': Index_Objective,
                  'Product_Objective': Product_Objective,
                  'Divide_Objective': Divide_Objective}



"""Preset parameter configurations for unit testing"""

from obsidian.parameters import Param_Continuous, Param_Ordinal, Param_Categorical, \
    Param_Observational, Param_Discrete_Numeric, ParamSpace

# Set up ap master list of parameter spaces for testing
params = [
        Param_Continuous('Parameter 1', 0, 10),
        Param_Continuous('Parameter 2', -20, 0),
        Param_Continuous('Parameter 3', 5, 5),
        Param_Observational('Parameter 4', 0, 24),
        Param_Discrete_Numeric('Parameter 5', [-2, -1, 1, 2]),
        Param_Discrete_Numeric('Parameter 6', [7]),
        Param_Categorical('Parameter 7', ['A', 'B', 'C', 'D']),
        Param_Categorical('Parameter 8', ['E', 'F', 'G', 'H']),
        Param_Categorical('Parameter 9', ['I']),
        Param_Ordinal('Parameter 10', ['J', 'K', 'L', 'M']),
        Param_Ordinal('Parameter 11', ['N'])
    ]

# Subset some default selections
default = [params[i] for i in [0, 1, 2, 6]]  # 2 continuous, 1 static, 1 categorical
X_sp_default = ParamSpace(params=default)

# Numeric
cont_small = [params[i] for i in [0, 1, 2]]  # continuous including edge cases
numeric = [params[i] for i in [0, 1, 2, 3, 4, 5]]  # numeric including edge cases
X_sp_cont_small = ParamSpace(params=cont_small)
X_sp_numeric = ParamSpace(params=numeric)

# Nominal
cat_small = [params[i] for i in [6, 7, 8]]  # categorical including edge cases
disc_small = [params[i] for i in [6, 9]]  # 1 categorical, 1 ordinal
disc_large = [params[i] for i in [6, 7, 8, 9, 10]]  # discrete including edge cases
X_sp_cat_small = ParamSpace(params=cat_small)
X_sp_disc_small = ParamSpace(params=disc_small)
X_sp_disc_large = ParamSpace(params=disc_large)

# Set up a range of continuous parameters
params_cont_large = [
        Param_Continuous('Parameter 1', 0, 10),
        Param_Continuous('Parameter 2', 0, 10),
        Param_Continuous('Parameter 3', 0, 10),
        Param_Continuous('Parameter 4', 0, 10),
        Param_Continuous('Parameter 5', 0, 10),
        Param_Continuous('Parameter 6', 0, 10),
        Param_Continuous('Parameter 7', 0, 10),
        Param_Continuous('Parameter 8', 0, 10),
        Param_Continuous('Parameter 9', 0, 10),
        Param_Continuous('Parameter 10', 0, 10),
        Param_Continuous('Parameter 11', 0, 10),
        Param_Continuous('Parameter 12', 0, 10),
]
X_sp_cont_large = ParamSpace(params=params_cont_large)
X_sp_cont_ndims = [ParamSpace(params_cont_large[:i]) for i in range(len(params_cont_large))]

# Wrap everything for iteration during testing
test_X_space = [X_sp_default, X_sp_cont_small, X_sp_numeric, X_sp_cat_small, X_sp_disc_small, X_sp_disc_large]



"""Output features to be optimized"""

import pandas as pd
import numpy as np
import torch
from .transforms import f_transform_dict
from numpy.typing import ArrayLike

from obsidian.exceptions import UnfitError


class Target():
    """
    Base class for optimization response targets.
    """

    def __init__(self,
                 name: str,
                 f_transform: str | None = 'Standard',
                 aim: str = 'max'):

        self.name = name
        if aim not in ['min', 'max']:
            raise ValueError('Aim must be either "min" or "max"')
        if aim == 'min':
            self.multiplier = -1
        else:
            self.multiplier = 1
        self.aim = aim
        
        # Ouput scoring, used for transformation OR to create a cost function of multiple outputs/inputs
        if f_transform is not None:
            if f_transform not in f_transform_dict.keys():
                raise KeyError(f'Scoring function must be selected from one of: {f_transform_dict.keys()}')
        else:
            f_transform = 'Identity'
        self.f_transform = f_transform

    def __repr__(self):
        """String representation of object"""
        return f"{self.__class__.__name__}({self.name}, aim={self.aim})"

    def transform_f(self,
                    f: float | int | ArrayLike,
                    inverse=False,
                    fit=False):
        """
        Converts a raw response to an objective function value ("score").
        Cost-penalization and response transformation should be handled here.

        Args:
            f (array-like): The column(s) containing the response values (y)
            inverse (bool, optional): An indicator to perform the inverse transform. Defaults to ``False``.
            fit (bool, optional): An indicator to fit the properties of the transform function. Defaults to ``False``.

        Returns:
            pd.Series: An array of transformed f values matching the responses in Z

        Raises:
            TypeError: If f is not numeric or array-like
            UnfitError: If the transform function is called without being fit first
        """

        if not (isinstance(f, (pd.Series, pd.DataFrame, np.ndarray, list, float, int))
                or torch.is_tensor(f)):
            raise TypeError('f being transformed must be numeric or array-like')
        
        # Convert everything to numpy except Tensors
        if isinstance(f, (float, int)):
            f = np.array([f])
        if isinstance(f, (list)):
            f = np.array(f)
        if isinstance(f, (pd.Series, pd.DataFrame)):
            f = f.values

        if not torch.is_tensor(f):
            # Check that types are valid, then convert to Tensor
            if not all(np.issubdtype(f_i.dtype, np.number) for f_i in f.flatten()):
                raise TypeError('Each element of f being transformed must be numeric')
            f = torch.tensor(f)

        if not fit:
            if not hasattr(self, 'f_transform_func'):
                raise UnfitError('Transform function is being called without being fit first.')

        if f.ndim == 1:
            f = f.reshape(-1, 1)

        if inverse:
            f_obj = self.f_transform_func.inverse(f * self.multiplier)
            return pd.Series(f_obj.flatten(), name=self.name)
        else:
            if fit:
                self.f_transform_func = f_transform_dict[self.f_transform]()
                f_obj = self.f_transform_func(f, fit=True)
                self.f_raw = f  # Save raw data for re-loading and re-fitting state as needed
            else:
                f_obj = self.f_transform_func(f)
            return pd.Series(f_obj.flatten(), name=self.name+' Trans') * self.multiplier
        
    def save_state(self) -> dict:
        """
        Saves the state of the object as a dictionary.

        Returns:
            dict: A dictionary containing the state of the object.
        """
        # Prepare a dictionary to describe the state
        obj_dict = {'init_attrs': {}}

        # Select some optimizer attributes to save directly
        init_attrs = ['name', 'aim', 'f_transform']
        for attr in init_attrs:
            obj_dict['init_attrs'][attr] = getattr(self, attr)

        # If the transformer has been fit, store the raw data so it can be refit upon load
        if hasattr(self, 'f_transform_func'):
            obj_dict['f_raw'] = self.f_raw.tolist()

        return obj_dict

    @classmethod
    def load_state(cls, obj_dict: dict):
        """
        Loads the state of the target object from a dictionary.

        Args:
            cls (class): The class of the target object.
            obj_dict (dict): A dictionary containing the state of the target object.

        Returns:
            The loaded target object.
        """
        new_target = cls(**obj_dict['init_attrs'])

        # If the transformer has been fit before saving, refit it
        f = torch.Tensor(obj_dict['f_raw'])
        new_target.transform_f(f, fit=True)

        return new_target



from .utils import add_tab, make_input, make_dropdown, make_switch, make_slider, make_knob, make_table, make_collapse
from .utils import center
import dash_bootstrap_components as dbc
from dash import dcc, html, Dash, dash_table, callback, Output, Input, State, ALL, MATCH
import pandas as pd
import base64
import io
from dash.exceptions import PreventUpdate


def make_acquisition(index, delete=True):
    
    aq_functions = [{'label': 'Expected Improvement', 'value': 'EI'},
                    {'label': 'Probability of Improvement', 'value': 'PI'},
                    {'label': 'Upper Confidence Band', 'value': 'UCB'},
                    {'label': 'Space Filling', 'value': 'SF'}]
    columns = [dbc.Col(make_dropdown('', 'Select an acquisition function', aq_functions,
                                     id={'type': 'input-acquisition', 'index': index},
                                     kwargs={'value': aq_functions[0]['value']}),
                       width=7 if delete else 9)]
    
    hyper_input = [dbc.Input(id={'type': 'input-alpha', 'index': index})]
    if delete:
        hyper_input.append(dbc.Button('Delete', id={'type': 'aq_delete', 'index': index},
                                      className='me-2', color='danger', n_clicks=0))

    columns.append(dbc.Col([dbc.InputGroup(hyper_input),
                            dbc.FormText('Input a hyperparameter (optional)', style={'font-size': '0.7em'})]))

    return dbc.Row(id={'type': 'div-acquisition', 'index': index}, children=columns, align='center')


def setup_config(app, app_tabs):

    input_optimizer_seed = make_input('Optimizer Seed', 'Randomization seed for Bayesian optimizer', required=False,
                                      id='input-optimizer_seed',
                                      kwargs={'type': 'number', 'min': 1, 'max': 99999,
                                              'step': 1, 'placeholder': 'None'})
    
    options_f_transform = ['Standard', 'Power']
    
    input_f_transform = make_dropdown('Response Transform', 'Select transformation for response', options_f_transform,
                                      id='input-f_transform', kwargs={'value': options_f_transform[0]})
    
    options_surrogate = [{'label': 'Default Gaussian Process', 'value': 'GP'},
                         {'label': 'Uninformed Gaussian Process', 'value': 'GPflat'},
                         {'label': 'Gaussian Process with Custom Prior', 'value': 'GPprior'},
                         {'label': 'Deep Kernel Learning Gaussian Process', 'value': 'DKL'}]
    
    input_surrogate = make_dropdown('Surrogate Model', 'Select a surrogate model', options_surrogate,
                                    id='input-surrogate', kwargs={'value': options_surrogate[0]['value']})
    
    input_m_batch = make_input('Number of Suggestions', 'The number of suggestions for the \
                               optimizer to provide (per acquisition function)', 1,
                               id='input-m_batch', kwargs={'type': 'number', 'min': 1, 'max': 16, 'step': 1})
    input_optim_sequential = make_switch('Optimize Sequentially', 'Switch for optimizing batch \
                                         suggestions sequentially or simultaneously', id='toggle-optim_sequential')
    input_optim_restarts = make_slider('Optimizer Restarts', 'Number of automatic restarts for \
                                        acquisition optimization', 1, 100,
                                       id='input-optim_restarts', kwargs={'value': 10})

    acquisitions = dbc.Container(dbc.Card(id='aq_inputs', children=[
        dbc.CardHeader(['Acquisition Functions', html.Div(dbc.FormText('Objective function of the experiment\
                                                                      search/selection',
                                                          style={'font-size': '0.7em'}))]),
        dbc.CardBody(html.Div(id='div-acquisition_all', children=[make_acquisition(index=0, delete=False)])),
        dbc.CardFooter(dbc.Button('Add', id='button-acquisition_add', className='me-2', color='secondary', n_clicks=0))
        ], style={'margin-top': '15px'}))
    
    general_options = dbc.Container(dbc.Card(dbc.CardBody(
        children=[input_surrogate, input_m_batch, input_optim_sequential])),
                                    style={'margin-top': '15px'})
    
    advanced_options = dbc.Container(make_collapse('adv_options',
                                                   [input_optimizer_seed, input_f_transform, input_optim_restarts],
                                                   'Advanced Options'),
                                     style={'margin-top': '15px'})

    # Config store
    config = dcc.Store(id='store-config')
    
    # Extra div for printing outputs, for troubleshooting
    troubleshoot = html.Div(id='debug-print-config')
    
    elements = [html.Br(), general_options, acquisitions, advanced_options, config, html.Hr(), troubleshoot]
    add_tab(app_tabs, elements, 'tab-config', 'Config')
    setup_config_callbacks(app)
    
    return


def setup_config_callbacks(app):
    
    @app.callback(
        Output('collapse-adv_options', 'is_open'),
        Input('button-collapse-adv_options', 'n_clicks'),
        State('collapse-adv_options', 'is_open'),
    )
    def toggle_collapse(n, is_open):
        if n:
            return not is_open
        return is_open
    
    # Add acquisition function
    @app.callback(
        Output('div-acquisition_all', 'children'),
        Output('button-acquisition_add', 'n_clicks'),
        Input('button-acquisition_add', 'n_clicks'),
        State('div-acquisition_all', 'children'),
    )
    def add_aq(add_clicked, aq_funs):
        if add_clicked:
            aq_funs.append(make_acquisition(index=len(aq_funs)+1))
        return aq_funs, 0
    
    # Delete acquisition function
    @app.callback(
        Output({'type': 'div-acquisition', 'index': MATCH}, 'children'),
        Output({'type': 'aq_delete', 'index': MATCH}, 'n_clicks'),
        Input({'type': 'aq_delete', 'index': MATCH}, 'n_clicks'),
        State({'type': 'div-acquisition', 'index': MATCH}, 'children'),
    )
    def func(del_clicked, aq_match):
        if del_clicked:
            aq_match = None
        return aq_match, 0
    
    # Store all of the input and settings into config
    @app.callback(
        Output('store-config', 'data'),
        Input('input-response_name', 'value'),
        Input('input-optimizer_seed', 'value'),
        Input('input-f_transform', 'value'),
        Input('input-surrogate', 'value'),
        Input('input-m_batch', 'value'),
        Input('toggle-optim_sequential', 'value'),
        Input('input-optim_restarts', 'value'),
        Input({'type': 'input-acquisition', 'index': ALL}, 'value'),
        Input({'type': 'input-alpha', 'index': ALL}, 'value'),
        prevent_initial_call=True
    )
    def compile_config(response_name, optimizer_seed, f_transform, surrogate,
                       m_batch, optim_sequential, optim_restarts, aq, alpha):
        
        config = {}
        
        config['response_name'] = response_name
        config['optimizer_seed'] = int(optimizer_seed) if optimizer_seed is not None else None
        config['surrogate_params'] = {'f_transform': f_transform, 'surrogate': surrogate}
        # TODO: Re-implement hyperparmeters based on selection in alpha
        config['aq_params'] = {'optim_sequential': optim_sequential, 'optim_restarts': optim_restarts,
                               'm_batch': m_batch, 'acquisition': aq}
        config['verbose'] = 0
        
        return config
    
    @app.callback(
        Output('debug-print-config', 'children'),
        Input('store-config', 'data'),
    )
    def troubleshoot_config(config):
        return
    
    return



"""Parameters that can only be sampled at specific values"""

from .base import Parameter
from .utils import transform_with_type

from obsidian.config import CAT_SEP

from numpy.typing import ArrayLike

import pandas as pd
import numpy as np


class Param_Discrete(Parameter):
    """
    Represents a discrete parameter.

    Attributes:
        name (str): The name of the parameter.
        categories (list[str]): The categories of the parameter.

    Properties:
        min (int): The minimum value of the parameter (always 0).
        nc (int): The number of categories.
        max (int): The maximum value of the parameter (nc - 1).

    """

    @transform_with_type
    def unit_map(self, X: np.ndarray):
        """
        Maps the input values to the unit interval [0, 1] based on the categories.

        Parameters:
            X (np.ndarray): The input values to be mapped.

        Returns:
            np.ndarray: Values mapped to the unit interval [0, 1].
        """
        X_str = X.flatten().astype('U32')
        indices = [self.categories.index(x) for x in X_str]
        return (np.array(indices)/self.nc).reshape(X.shape)
    
    @transform_with_type
    def unit_demap(self, X: np.ndarray):
        """
        Maps continuous values to discrete categories.

        Parameters:
            X (np.ndarray): The input values in the interval [0,1]

        Returns:
            np.ndarray: Values mapped back to the real space.
        """

        return np.array(self.categories)[(X.flatten()*self.nc)
                                         .astype('int')].reshape(X.shape).astype('U32')

    @property
    def min(self):
        """Minimum parameter value (always 0 for discrete)"""
        return 0

    @property
    def nc(self):
        """Number of discrete categories"""
        return len(self.categories)
        
    @property
    def max(self):
        """Maximum parameter value (nc-1)"""
        return self.nc-1
        
    def set_search(self,
                   search_categories: list[str]):
        """
        Set the search space for the parameter

        Args:
            search_categories (list[str]): The search space for the parameter.
        """
        for c in search_categories:
            self._validate_value(c)
            
        self.search_categories = search_categories
    
    def open_search(self):
        """Set the search space to the parameter space"""
        self.set_search(self.categories)
    
    def _validate_value(self,
                        value: str):
        """
        Validates if the given value is in the list of categories.

        Args:
            value (str): The value to be validated.

        Raises:
            KeyError: If the value is not in the list of categories.
            TypeError: If the value is not a string
        """
        if len(value) > 32:
            raise ValueError(f'Category names must be <= 32 characters. Too long value encountered: {value}')

        if not isinstance(value, str):
            raise TypeError(f'Value {value} is not a string')
        
        if value not in self.categories:
            raise ValueError(f'Value {value} is not in {self.categories}')

    def __init__(self,
                 name: str,
                 categories: str | list[str],
                 search_categories: list[str] = None):
        super().__init__(name=name)
        if isinstance(categories, str):
            self.categories = categories.split(',')
            self.categories = [c.rstrip().lstrip() for c in self.categories]
        else:
            self.categories = categories
        for c in self.categories:
            self._validate_value(c)
        
        # Set the search space to the parameter space by default
        if not search_categories:
            search_categories = self.categories
        else:
            if isinstance(categories, str):
                search_categories = search_categories.split(',')
                search_categories = [c.rstrip().lstrip() for c in search_categories]
        self.set_search(search_categories)
    
    def __repr__(self):
        """String representation of object"""
        return f"{self.__class__.__name__}(name={self.name}, categories={self.categories})"


class Param_Ordinal(Param_Discrete):
    """
    Represents an ordinal parameter; a discrete parameter with an order.
    """
    # Ordinal encoder maps to integers

    def encode(self, X: np.ndarray):
        """Encode parameter to a format that can be used for training"""
        return self.unit_map(X)
    
    def decode(self, X: np.ndarray):
        """Decode parameter from transformed space"""
        return self.unit_demap(X)


class Param_Categorical(Param_Discrete):
    """
    Represents an categorical parameter; a discrete parameter without an order.
    """

    # Categorical encoder maps to one-hot columns
    # No decorator on this, we need to handle dataframes
    def encode(self, X: str | ArrayLike):
        """Encode parameter to a format that can be used for training"""
        X_str = np.array(X).flatten().astype('U32')
        X_cat = pd.Series(X_str).astype(pd.CategoricalDtype(categories=self.categories))
        X_ohe = pd.get_dummies(X_cat, prefix_sep=CAT_SEP, dtype=float, prefix=self.name)

        return X_ohe
    
    def decode(self, X: ArrayLike):
        """Decode parameter from transformed space"""
        return [self.categories[int(x)] for x in np.array(X).argmax(axis=1)]


class Task(Param_Discrete):
    """
    Represents an task parameter; a discrete parameter indicating a distinct system.
    """
    # Similar to ordinal, but (0,nc-1) instead of (0,1)

    @transform_with_type
    def encode(self, X: np.ndarray):
        """Encode parameter to a format that can be used for training"""
        return self.unit_map(X)*self.nc
    
    @transform_with_type
    def decode(self, X: np.ndarray):
        """Decode parameter from transformed space"""
        return np.array(self.unit_demap(X/self.nc)).astype('U32')


class Param_Discrete_Numeric(Param_Discrete):
    """
    Represents an discrete numeric parameter; an ordinal parameter comprised of numbers.

    Raises:
        TypeError: If the categories are not numbers.
        TypeError: If the categories are not a list of numbers.
    """

    @property
    def min(self):
        """Minimum parameter value"""
        return np.array(self.categories).min()

    @property
    def max(self):
        """Maximum parameter value"""
        return np.array(self.categories).max()
    
    @property
    def range(self):
        """The range of the parameter (max - min)"""
        return self.max-self.min

    def __init__(self,
                 name,
                 categories: list[int | float],
                 search_categories: list[int | float] = None):
        
        if not isinstance(categories, list):
            raise TypeError('Categories must be a number or list of numbers')

        self.name = name
        self.categories = categories if isinstance(categories, list) else [categories]
        self.categories.sort()
        for c in self.categories:
            self._validate_value(c)
        
        # Set the search space to the parameter space by default
        if not search_categories:
            search_categories = self.categories
        else:
            if not isinstance(search_categories, list):
                search_categories = [search_categories]
            for c in search_categories:
                self._validate_value(c)
        search_categories.sort()
        self.set_search(search_categories)
    
    def _validate_value(self,
                        value: int | float):
        """
        Validates if the given value is in the list of categories.

        Args:
            value (str): The value to be validated.

        Raises:
            KeyError: If the value is not in the list of categories.
            TypeError: If the value is not a string
        """
        if not isinstance(value, (int, float)):
            raise TypeError(f'Value {value} is not numeric')
        
        if value not in self.categories:
            raise ValueError(f'Value {value} is not in {self.categories}')

    @transform_with_type
    def unit_map(self, X: np.ndarray):
        """Map from measured to 0,1 space"""
        return (X-self.min)/self.range if self.range != 0 else 0*X

    @transform_with_type
    def unit_demap(self, X: np.ndarray):
        """Map from 0,1 to measured space"""
        closest_idx = np.abs(np.array(self.categories)
                             - (X.flatten()[..., np.newaxis]*self.range+self.min)).argmin(axis=1)
        return np.array(self.categories)[closest_idx].reshape(X.shape)

    encode = unit_map
    decode = unit_demap
